import copy
import hashlib
import re
from contextlib import closing
from typing import Any, Callable, Iterator, List, Literal, Optional, Tuple, Union, cast

import numpy
import pandas
import pyarrow
from sqlalchemy.sql.compiler import RESERVED_WORDS

from ..catalog.orm import Node
from ..storage import EmbeddedSQLStorage, RemoteSQLStorage, SQLStorage, get_storage
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource
from ..structures.table import TableStructure
from ..type_aliases import JSON
from .array import ArrayAdapter
from .utils import init_adapter_from_catalog

DIALECTS = Literal["postgresql", "sqlite", "duckdb"]
TABLE_NAME_PATTERN = re.compile(r"^[a-z][a-z0-9_]*$")
COLUMN_NAME_PATTERN = re.compile(r"^[a-zA-Z_].*$")
FORBIDDEN_CHARACTERS = re.compile(
    r"""
    (
      '           # single quote -- ends string literals
    | "           # double quote -- ends quoted identifiers (PostgreSQL)
    | `           # backtick -- ends quoted identifiers (MySQL)
    | ;           # semicolon -- terminates statements (SQL injection risk)
    | --          # double dash -- starts comments (SQL injection risk)
    | /\*         # \* -- starts block comment
    | \*/         # */ -- ends block comment
    | \\          # backslash -- escape character (esp. in MySQL)
    | %           # percent -- wildcards (in LIKE patterns, can be tricky if misused)
    | \(          # parenthesis -- alters grouping, can break expressions
    | \)          #
    | =           # equals -- can change logic (injections in expressions)
    | \+          # plus -- can change logic (injections in expressions)
    )
    """,
    re.VERBOSE,
)
# NOTE: While capital letters are allowed in SQL identifiers, we convert column names
# to lower case to avoid potential collisions (in SQLite). The original name with
# upper case letters is retained in the structure, but attempts to create columns
# e.g. "A" and "a" will raise an error.
# Furthermore, user-specified table names can only be in lower case.


class SQLAdapter:
    """SQLAdapter Class

    This class provides an interface for interacting with SQL databases.

    Parameters
    ----------
    data_uri : the uri of the database, starting either with "duckdb://" or "postgresql://"
    structure : the structure of the data; structure is not optional for sql database
    table_name : the name of the table in the database. Will be converted to lower case in
        all SQL queries.
    dataset_id : the dataset id of the data in the storage database.
    metadata : the optional metadata of the data.
    specs : the specs.
    """

    structure_family = StructureFamily.table
    supported_storage = {SQLStorage, EmbeddedSQLStorage, RemoteSQLStorage}

    def __init__(
        self,
        data_uri: str,
        structure: TableStructure,
        table_name: str,
        dataset_id: int,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        self.storage: SQLStorage = cast(SQLStorage, get_storage(data_uri))
        self._metadata = metadata or {}
        self._structure = structure
        self.specs = list(specs or [])
        self.table_name = table_name
        self.dataset_id = dataset_id

    def metadata(self) -> JSON:
        """The metadata representing the actual data.

        Returns
        -------
        The metadata representing the actual data.
        """
        return self._metadata

    @classmethod
    def init_storage(
        cls,
        storage: SQLStorage,
        data_source: DataSource[TableStructure],
        path_parts: Optional[List[str]] = None,
    ) -> DataSource[TableStructure]:
        """
        Class to initialize the list of assets for given uri.

        Parameters
        ----------
        storage: SQLStorage
        data_source : DataSource
        path_parts : List[str]
            Not used by this adapter

        Returns
        -------
        A modified copy of the data source
        """
        data_source = copy.deepcopy(data_source)  # Do not mutate caller input.

        # Create a table_name based on the hash of Arrow schema
        schema = data_source.structure.arrow_schema_decoded
        encoded = schema.serialize()
        default_table_name = "table_" + hashlib.md5(encoded).hexdigest().lower()
        table_name = data_source.parameters.setdefault("table_name", default_table_name)
        is_safe_identifier(table_name, TABLE_NAME_PATTERN, allow_reserved_words=False)

        # Prefix columns with internal _dataset_id, _partition_id, ...
        schema = schema.insert(0, pyarrow.field("_partition_id", pyarrow.int16()))
        schema = schema.insert(0, pyarrow.field("_dataset_id", pyarrow.int32()))
        create_table_statement = arrow_schema_to_create_table(
            schema, table_name, cast(DIALECTS, storage.dialect)
        )

        create_index_statement = (
            "CREATE INDEX IF NOT EXISTS dataset_and_partition_index "
            f'ON "{table_name}"(_dataset_id, _partition_id)'
        )

        with closing(storage.connect()) as conn:
            with conn.cursor() as cursor:
                cursor.execute(create_table_statement)
            with conn.cursor() as cursor:
                cursor.execute(create_index_statement)
            # Just once, create a SEQUENCE (or the closest analogue in SQLite) to
            # provide unique dataset_id for each dataset in this database.
            # (If it exists, do nothing.)
            # Then obtain the next value in the SEQUENCE for the dataset we are
            # initializing here.
            if storage.dialect == "sqlite":
                # Create single-row table with a counter, if it does not exist.
                with conn.cursor() as cursor:
                    cursor.execute(
                        "CREATE TABLE IF NOT EXISTS _dataset_id_counter (value INTEGER NOT NULL)"
                    )
                with conn.cursor() as cursor:
                    cursor.execute(
                        "INSERT INTO _dataset_id_counter (value) "
                        "SELECT 0 WHERE NOT EXISTS (SELECT 1 FROM _dataset_id_counter)"
                    )
                with conn.cursor() as cursor:
                    # Increment the counter.
                    cursor.execute(
                        "UPDATE _dataset_id_counter SET value = value + 1 "
                        "RETURNING value"
                    )
                    (dataset_id,) = cursor.fetchone()
            else:
                with conn.cursor() as cursor:
                    cursor.execute("CREATE SEQUENCE IF NOT EXISTS _dataset_id_counter")
                with conn.cursor() as cursor:
                    cursor.execute("SELECT nextval('_dataset_id_counter')")
                    (dataset_id,) = cursor.fetchone()
            data_source.parameters["dataset_id"] = dataset_id
            conn.commit()

        data_source.assets.append(
            Asset(
                # Store URI *without* credentials.
                data_uri=storage.uri,
                is_directory=False,
                parameter="data_uri",
                num=None,
            )
        )
        return data_source

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[TableStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "SQLAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    def structure(self) -> TableStructure:
        """
        The structure of the actual data.

        Returns
        -------
        The structure of the data.
        """
        return self._structure

    def get(self, key: str) -> Union[ArrayAdapter, None]:
        """Get the data for a specific key

        Parameters
        ----------
        key : a string to indicate which column to be retrieved

        Returns
        -------
        The column for the associated key.
        """
        if key not in self.structure().columns:
            return None
        return self[key]

    def __getitem__(self, key: str) -> ArrayAdapter:
        """Get the data for a specific key.

        Parameters
        ----------
        key : a string to indicate which column to be retrieved

        Returns
        -------
        The column for the associated key.
        """

        # Must compute to determine shape.
        return ArrayAdapter.from_array(self.read([key])[key].values)

    def items(self) -> Iterator[Tuple[str, ArrayAdapter]]:
        """Iterate over the SQLAdapter data.

        Returns
        -------
        An iterator for the data in the associated database.
        """
        yield from ((key, self[key]) for key in self._structure.columns)

    def append_partition(
        self,
        data: Union[
            List[pyarrow.record_batch],
            pyarrow.record_batch,
            pandas.DataFrame,
            pyarrow.Table,
        ],
        partition: int,
    ) -> None:
        """Write the data as arrow format

        Parameters
        ----------
        data : data to append to the database. Can be a record_batch, a list of record batches, pyarrow table, or a
            pandas dataframe.
        partition : the partition index to write.
        """
        # Convert the data to pyarrow table
        if isinstance(data, pyarrow.Table):
            table = data
        elif isinstance(data, pandas.DataFrame):
            table = pyarrow.Table.from_pandas(data)
        else:
            if not isinstance(data, list):
                batches = [data]
            else:
                batches = data
            table = pyarrow.Table.from_batches(batches)
        # Explicitly cast the table to the structure's schema
        table = table.cast(self.structure().arrow_schema_decoded)
        # Prepend columns for internal dataset_id and partition number.
        dataset_id_column = self.dataset_id * numpy.ones(len(table), dtype=numpy.int32)
        partition_id_column = partition * numpy.ones(len(table), dtype=numpy.int16)
        table = table.add_column(
            0, pyarrow.field("_partition_id", pyarrow.int16()), [partition_id_column]
        )
        table = table.add_column(
            0, pyarrow.field("_dataset_id", pyarrow.int32()), [dataset_id_column]
        )
        # Convert column names to lower case
        if upr_lwr_case_mapping := {
            c: c.lower() for c in table.column_names if c != c.lower()
        }:
            table = table.rename_columns(upr_lwr_case_mapping)

        with closing(self.storage.connect()) as conn:
            with conn.cursor() as cursor:
                cursor.adbc_ingest(self.table_name, table, mode="append")
            conn.commit()

    def _read_full_table_or_partition(
        self, fields: Optional[List[str]] = None, partition: Optional[int] = None
    ) -> pyarrow.Table:
        """Read the data from the database

        This is a helper function to read the data from the database. The result
        is a pyarrow table containing rows either from the entire table or from a
        specific partition. The retained columns are cast to the original type.

        Parameters
        ----------
        fields : optional string to return the data in the specified field.
        partition : optional int to return the data in the specified partition.

        Returns
        -------
        The concatenated table as pyarrow table.
        """

        # Make sure that requested columns exist and safe to put in SQL query.
        schema = self.structure().arrow_schema_decoded
        req_cols = set(schema.names).intersection(fields) if fields else schema.names

        query = (
            "SELECT " + ", ".join([f'"{c.lower()}"' for c in req_cols]) + " "
            f'FROM "{self.table_name}" '
            f"WHERE _dataset_id={self.dataset_id} "
        )
        query += (
            f"AND _partition_id={int(partition)}"
            if partition is not None
            else "ORDER BY _partition_id"
        )

        with closing(self.storage.connect()) as conn:
            with conn.cursor() as cursor:
                cursor.execute(query)
                data = cursor.fetch_arrow_table()
            conn.commit()

        # The database may have stored this in a coarser type, such as
        # storing uint8 data as int16. Cast it to the original type.
        # Additionally, back-convert lower case column names to their original names.
        if lwr_upr_case_mapping := {c.lower(): c for c in req_cols if c != c.lower()}:
            data = data.rename_columns(lwr_upr_case_mapping)
        target_schema = pyarrow.schema(
            [schema.field(schema.get_field_index(name)) for name in req_cols]
        )

        return data.cast(target_schema)

    def read(self, fields: Optional[List[str]] = None) -> pandas.DataFrame:
        """Read the concatenated data from the entire table.

        Parameters
        ----------
        fields: optional string to return the data in the specified field.

        Returns
        -------
        The concatenated table as pandas dataframe.
        """

        return self._read_full_table_or_partition(fields=fields).to_pandas()

    def read_partition(
        self, partition: int, fields: Optional[List[str]] = None
    ) -> pandas.DataFrame:
        """Read a batch of data from a given partition.

        Parameters
        ----------
        partition : int
        fields : Optional[List[str]]
            Optional list of field names to select. By default return all.

        Returns
        -------
        The concatenated table as pandas dataframe.
        """

        return self._read_full_table_or_partition(
            fields=fields, partition=partition
        ).to_pandas()


# Mapping between Arrow types and PostgreSQL column type name.
ARROW_TO_PG_TYPES: dict[pyarrow.Field, str] = {
    # Boolean
    pyarrow.bool_(): "BOOLEAN",
    # Integers
    pyarrow.int8(): "SMALLINT",  # no native int8, use int16
    pyarrow.int16(): "SMALLINT",
    pyarrow.int32(): "INTEGER",
    pyarrow.int64(): "BIGINT",
    # Since PG has not native unsigned int,
    # use signed int one size up.
    pyarrow.uint8(): "SMALLINT",
    pyarrow.uint16(): "INTEGER",
    pyarrow.uint32(): "BIGINT",
    # pyarrow.uint64() not supported
    # Floating Point
    # pyarrow.float16(): "REAL",
    pyarrow.float32(): "REAL",
    pyarrow.float64(): "DOUBLE PRECISION",
    # String Types
    pyarrow.string(): "TEXT",
    pyarrow.large_string(): "TEXT",
    # TODO Consider adding support for these types, with testing.
    # # Binary Types
    # pyarrow.binary(): "bytea",
    # pyarrow.large_binary(): "bytea",
    # # Date/Time Types
    # pyarrow.date32(): "date",
    # pyarrow.date64(): "date",
    # pyarrow.time32("s"): "time",
    # pyarrow.time64("us"): "time",
    # pyarrow.timestamp(
    #     "us"
    # ): "timestamp",  # Note: becomes timestamptz if timezone specified
    # # Duration/Interval
    # pyarrow.duration("us"): "interval",
    # # Decimal Types
    # pyarrow.decimal128(precision=38, scale=9): "numeric",
    # pyarrow.decimal256(precision=76, scale=18): "numeric",
}


def arrow_field_to_pg_type(field: Union[pyarrow.Field, pyarrow.DataType]) -> str:
    """Get the PostgreSQL type name for a given PyArrow field.

    Parameters
    ----------
    field : pyarrow.Field
        The PyArrow field containing type and metadata information

    Returns
    -------
    str
        The corresponding PostgreSQL type name

    Raises
    ------
    ValueError
        If the field's type is not supported

    Examples
    --------
    >>> import pyarrow as pa
    >>> field = pyarrow.field('timestamp_col', pyarrow.timestamp('us', tz='UTC'))
    >>> get_pg_type(field)
    'timestamptz'
    >>> field = pyarrow.field('int_list', pyarrow.list_(pyarrow.int32()))
    >>> get_pg_type(field)
    'int4 ARRAY'
    >>> field = pyarrow.field('dict_col', pyarrow.dictionary(pyarrow.int8(), pyarrow.string()))
    >>> get_pg_type(field)
    'text'
    """

    def _resolve_type(
        arrow_type: Union[pyarrow.Field, pyarrow.DataType]
    ) -> Union[pyarrow.Field, pyarrow.DataType, str]:
        """Internal helper to resolve type, handling nested cases."""
        # Handle list types
        if pyarrow.types.is_list(arrow_type):
            value_type = _resolve_type(arrow_type.value_type)
            return f"{value_type} ARRAY"

        # Handle fixed size list types
        if pyarrow.types.is_fixed_size_list(arrow_type):
            value_type = _resolve_type(arrow_type.value_type)
            return f"{value_type} ARRAY"

        # Handle large list types
        if pyarrow.types.is_large_list(arrow_type):
            value_type = _resolve_type(arrow_type.value_type)
            return f"{value_type} ARRAY"

        # TODO Consider adding support for these types, with testing.

        # # Handle dictionary types - use value type
        # if pyarrow.types.is_dictionary(arrow_type):
        #     return _resolve_type(arrow_type.value_type)

        # # Handle timestamp with timezone
        # if pyarrow.types.is_timestamp(arrow_type) and arrow_type.tz is not None:
        #     return "timestamptz"

        # Look up base type
        pg_type = ARROW_TO_PG_TYPES.get(arrow_type)
        if pg_type is None:
            raise ValueError(f"Unsupported PyArrow type: {arrow_type}")

        # TODO Consider adding support for these types, with testing.

        # # Special handling for time types with different units
        # if pyarrow.types.is_time(arrow_type):
        #     return "time"

        # # Special handling for timestamp types without timezone
        # if pyarrow.types.is_timestamp(arrow_type):
        #     return "timestamp"

        # # Special handling for duration/interval types with different units
        # if pyarrow.types.is_duration(arrow_type):
        #     return "interval"

        return pg_type

    return _resolve_type(field.type)


# Mapping between Arrow types and DuckDB column type names
ARROW_TO_DUCKDB_TYPES = {
    # Boolean
    pyarrow.bool_(): "BOOLEAN",
    # Integers
    pyarrow.int8(): "TINYINT",
    pyarrow.uint8(): "UTINYINT",
    pyarrow.int16(): "SMALLINT",
    pyarrow.uint16(): "USMALLINT",
    pyarrow.int32(): "INTEGER",
    pyarrow.uint32(): "UINTEGER",
    pyarrow.int64(): "BIGINT",
    pyarrow.uint64(): "UBIGINT",
    # Floating point
    # pyarrow.float16(): "REAL",  # Note: gets converted to float32 internally
    pyarrow.float32(): "REAL",
    pyarrow.float64(): "DOUBLE",
    # Decimal
    pyarrow.decimal128(precision=38, scale=9): "DECIMAL",
    pyarrow.decimal256(precision=76, scale=38): "DECIMAL",
    # String types
    pyarrow.string(): "VARCHAR",
    pyarrow.large_string(): "VARCHAR",
    # TODO Consider adding support for these types, with testing.
    # # Binary
    # pyarrow.binary(): 'BLOB',
    # pyarrow.large_binary(): 'BLOB',
    # # Temporal types
    # pyarrow.date32(): 'DATE',
    # pyarrow.date64(): 'DATE',
    # pyarrow.time32('s'): 'TIME',
    # pyarrow.time32('ms'): 'TIME',
    # pyarrow.time64('us'): 'TIME',
    # pyarrow.time64('ns'): 'TIME',
    # pyarrow.timestamp('s'): 'TIMESTAMP',
    # pyarrow.timestamp('ms'): 'TIMESTAMP',
    # pyarrow.timestamp('us'): 'TIMESTAMP',
    # pyarrow.timestamp('ns'): 'TIMESTAMP',
    # pyarrow.timestamp('us', tz='UTC'): 'TIMESTAMP WITH TIME ZONE',
    # # Interval/Duration
    # pyarrow.duration('s'): 'INTERVAL',
    # pyarrow.duration('ms'): 'INTERVAL',
    # pyarrow.duration('us'): 'INTERVAL',
    # pyarrow.duration('ns'): 'INTERVAL',
}


def arrow_field_to_duckdb_type(field: Union[pyarrow.Field, pyarrow.DataType]) -> str:
    """Get the DuckDB type name for a given PyArrow field.

    Parameters
    ----------
    field : Union[pyarrow.Field, pyarrow.DataType]
        The PyArrow field or type to convert

    Returns
    -------
    str
        The corresponding DuckDB type name

    Raises
    ------
    ValueError
        If the field's type is not supported or cannot be mapped

    Examples
    --------
    >>> import pyarrow as pa
    >>> get_duckdb_type(pyarrow.int32())
    'INTEGER'
    >>> get_duckdb_type(pyarrow.timestamp('us', tz='UTC'))
    'TIMESTAMP WITH TIME ZONE'
    >>> get_duckdb_type(pyarrow.list_(pyarrow.int32()))
    'INTEGER[]'
    >>> struct_type = pyarrow.struct([('x', pyarrow.int32()), ('y', pyarrow.string())])
    >>> get_duckdb_type(struct_type)
    'STRUCT(x INTEGER, y VARCHAR)'
    """

    def _resolve_type(
        arrow_type: Union[pyarrow.Field, pyarrow.DataType]
    ) -> Union[pyarrow.Field, pyarrow.DataType, str]:
        """Internal helper to resolve type, handling nested cases."""
        # Handle decimal types with custom precision/scale
        if pyarrow.types.is_decimal(arrow_type):
            return f"DECIMAL({arrow_type.precision}, {arrow_type.scale})"

        # Handle list types (including large lists and fixed size lists)
        if (
            pyarrow.types.is_list(arrow_type)
            or pyarrow.types.is_large_list(arrow_type)
            or pyarrow.types.is_fixed_size_list(arrow_type)
        ):
            value_type = _resolve_type(arrow_type.value_type)
            return f"{value_type}[]"

        # TODO Consider adding support for these types, with testing.

        # # Handle fixed size binary
        # if pyarrow.types.is_fixed_size_binary(arrow_type):
        #     return f'BLOB({arrow_type.byte_width})'

        # # Handle dictionary types - use value type
        # if pyarrow.types.is_dictionary(arrow_type):
        #     return _resolve_type(arrow_type.value_type)

        # # Handle timestamp with timezone
        # if pyarrow.types.is_timestamp(arrow_type):
        #     if arrow_type.tz is not None:
        #         return 'TIMESTAMP WITH TIME ZONE'
        #     return 'TIMESTAMP'

        # # Handle struct types
        # if pyarrow.types.is_struct(arrow_type):
        #     fields = []
        #     for field in arrow_type:
        #         field_type = _resolve_type(field.type)
        #         fields.append(f'{field.name} {field_type}')
        #     return f'STRUCT({", ".join(fields)})'

        # # Handle map types
        # if pyarrow.types.is_map(arrow_type):
        #     key_type = _resolve_type(arrow_type.key_type)
        #     item_type = _resolve_type(arrow_type.item_type)
        #     return f'MAP({key_type}, {item_type})'

        # Look up base type
        duckdb_type = ARROW_TO_DUCKDB_TYPES.get(arrow_type)
        if duckdb_type is None:
            raise ValueError(f"Unsupported PyArrow type: {arrow_type}")
        return duckdb_type

    arrow_type = field.type if isinstance(field, pyarrow.Field) else field
    return _resolve_type(arrow_type)


ARROW_TO_SQLITE_TYPES: dict[pyarrow.Field, str] = {
    # Boolean - stored as INTEGER
    pyarrow.bool_(): "INTEGER",
    # Integers - all stored as INTEGER
    pyarrow.int8(): "INTEGER",
    pyarrow.uint8(): "INTEGER",
    pyarrow.int16(): "INTEGER",
    pyarrow.uint16(): "INTEGER",
    pyarrow.int32(): "INTEGER",
    pyarrow.uint32(): "INTEGER",
    pyarrow.int64(): "INTEGER",
    # pyarrow.uint64() not supported, may exceed SQLite INTEGER range
    # Floating point - stored as REAL
    # pyarrow.float16(): "REAL",
    pyarrow.float32(): "REAL",
    pyarrow.float64(): "REAL",
    # Decimal - stored as TEXT to preserve precision
    pyarrow.decimal128(precision=38, scale=9): "TEXT",
    pyarrow.decimal256(precision=76, scale=38): "TEXT",
    # String types - stored as TEXT
    pyarrow.string(): "TEXT",
    pyarrow.large_string(): "TEXT",
    # TODO Consider adding support for these types, with testing.
    # # Binary - stored as BLOB
    # pyarrow.binary(): 'BLOB',
    # pyarrow.large_binary(): 'BLOB',
    # pyarrow.fixed_size_binary_type(32): 'BLOB',
    # # Temporal types - stored as TEXT or INTEGER
    # pyarrow.date32(): 'TEXT',  # ISO8601 date string
    # pyarrow.date64(): 'TEXT',  # ISO8601 date string
    # pyarrow.time32('s'): 'TEXT',  # ISO8601 time string
    # pyarrow.time64('us'): 'TEXT',  # ISO8601 time string
    # pyarrow.timestamp('s'): 'INTEGER',  # Unix timestamp
    # pyarrow.timestamp('ms'): 'INTEGER',  # Unix timestamp
    # pyarrow.timestamp('us'): 'INTEGER',  # Unix timestamp
    # pyarrow.timestamp('ns'): 'INTEGER',  # Unix timestamp
    # pyarrow.timestamp('us', tz='UTC'): 'TEXT',  # ISO8601 timestamp with TZ
    # # Duration/Interval - stored as INTEGER (microseconds) or TEXT
    # pyarrow.duration('s'): 'INTEGER',
    # pyarrow.duration('ms'): 'INTEGER',
    # pyarrow.duration('us'): 'INTEGER',
    # pyarrow.duration('ns'): 'INTEGER'
}


def arrow_field_to_sqlite_type(field: Union[pyarrow.Field, pyarrow.DataType]) -> str:
    """Get the SQLite type name for a given PyArrow field.

    Parameters
    ----------
    field : Union[pyarrow.Field, pyarrow.DataType]
        The PyArrow field or type to convert

    Returns
    -------
    str
        The corresponding SQLite type name

    Raises
    ------
    ValueError
        If the field's type is not supported or cannot be mapped

    Examples
    --------
    >>> import pyarrow as pa
    >>> get_sqlite_type(pyarrow.int32())
    'INTEGER'
    >>> get_sqlite_type(pyarrow.string())
    'TEXT'
    >>> get_sqlite_type(pyarrow.timestamp('us', tz='UTC'))
    'TEXT'
    >>> get_sqlite_type(pyarrow.list_(pyarrow.int32()))
    'TEXT'  # JSON encoded array
    """

    def _resolve_type(arrow_type: pyarrow.DataType) -> str:
        # TODO Consider adding support for these, with tests.

        # Handle dictionary types - use value type
        # if pyarrow.types.is_dictionary(arrow_type):
        #     return _resolve_type(arrow_type.value_type)

        # Handle timestamp with timezone - store as TEXT
        # if pyarrow.types.is_timestamp(arrow_type) and arrow_type.tz is not None:
        #     return "TEXT"

        # Handle nested types (lists, structs, maps) - store as JSON TEXT
        # if (
        #     pyarrow.types.is_list(arrow_type)
        #     or pyarrow.types.is_struct(arrow_type)
        #     or pyarrow.types.is_map(arrow_type)
        #     or pyarrow.types.is_fixed_size_list(arrow_type)
        #     or pyarrow.types.is_large_list(arrow_type)
        # ):
        #     return "TEXT"  # JSON encoded

        # Look up base type
        sqlite_type = ARROW_TO_SQLITE_TYPES.get(arrow_type)
        if sqlite_type is None:
            raise ValueError(f"Unsupported PyArrow type: {arrow_type}")
        return sqlite_type

    arrow_type = field.type if isinstance(field, pyarrow.Field) else field
    return _resolve_type(arrow_type)


DIALECT_TO_TYPE_CONVERTER: dict[
    DIALECTS, Callable[[Union[pyarrow.Field, pyarrow.DataType]], str]
] = {
    "duckdb": arrow_field_to_duckdb_type,
    "sqlite": arrow_field_to_sqlite_type,
    "postgresql": arrow_field_to_pg_type,
}


def arrow_schema_to_column_defns(
    schema: pyarrow.Schema, dialect: DIALECTS
) -> dict[str, str]:
    """
    Given Arrow schema, return mapping of column names to type definitions.

    Example output: {'x': 'INTEGER NOT NULL'}
    """
    # Check for possible column name collisions when converted to lower case
    all_names = [field.name.lower() for field in schema]
    if len(all_names) != len(set(all_names)):
        raise ValueError("Column names must be unique when converted to lower case.")
    columns = {}
    converter = DIALECT_TO_TYPE_CONVERTER[dialect]
    for field in schema:
        is_safe_identifier(field.name, COLUMN_NAME_PATTERN, allow_reserved_words=True)
        sql_type = converter(field)
        nullable = "NULL" if field.nullable else "NOT NULL"
        columns[field.name.lower()] = f"{sql_type} {nullable}"
    return columns


def arrow_schema_to_create_table(
    schema: pyarrow.Schema, table_name: str, dialect: DIALECTS
) -> str:
    "Construct the CREATE TABLE statement"
    columns = arrow_schema_to_column_defns(schema, dialect)
    return (
        f'CREATE TABLE IF NOT EXISTS "{table_name}" ('
        + ",\n ".join(f'"{name}" {type_}' for name, type_ in columns.items())
        + ")"
    )


def is_safe_identifier(
    identifier: str,
    pattern: re.Pattern[str],
    allow_reserved_words: bool = True,
) -> bool:
    if len(identifier) > 63:
        raise ValueError(
            f'Invalid SQL identifier "{identifier}": max character number is 63'
        )

    if not allow_reserved_words and identifier.lower() in RESERVED_WORDS:
        raise ValueError(
            f'Reserved SQL keywords are not allowed in identifiers, "{identifier}"'
        )

    if pattern.match(identifier) is None:
        raise ValueError(f'Malformed SQL identifier "{identifier}"')

    if match := FORBIDDEN_CHARACTERS.search(identifier):
        raise ValueError(
            f'Invalid SQL identifier "{identifier}" '
            f"contains forbidden character(s): {', '.join(match.groups())}"
        )

    return True



import collections
import dataclasses
import inspect
import os
import re
import warnings
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from functools import partial
from pathlib import Path
from typing import Callable, List, Optional, Set, TypeVar, Union

import anyio
import packaging
import pydantic_settings
from fastapi import (
    APIRouter,
    Body,
    Depends,
    HTTPException,
    Query,
    Request,
    Security,
    WebSocket,
)
from jmespath.exceptions import JMESPathError
from json_merge_patch import merge as apply_merge_patch
from jsonpatch import apply_patch as apply_json_patch
from starlette.requests import URL
from starlette.status import (
    HTTP_200_OK,
    HTTP_206_PARTIAL_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_405_METHOD_NOT_ALLOWED,
    HTTP_406_NOT_ACCEPTABLE,
    HTTP_410_GONE,
    HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from tiled.adapters.protocols import AnyAdapter
from tiled.media_type_registration import SerializationRegistry
from tiled.query_registration import QueryRegistry
from tiled.schemas import About
from tiled.server.protocols import ExternalAuthenticator, InternalAuthenticator
from tiled.server.schemas import Principal

from .. import __version__
from ..ndslice import NDSlice
from ..structures.core import Spec, StructureFamily
from ..type_aliases import Scopes
from ..utils import BrokenLink, ensure_awaitable, patch_mimetypes, path_from_uri
from ..validation_registration import ValidationError, ValidationRegistry
from . import schemas
from .authentication import (
    check_scopes,
    get_current_access_tags,
    get_current_access_tags_websocket,
    get_current_principal,
    get_current_principal_websocket,
    get_current_scopes,
    get_current_scopes_websocket,
    get_session_state,
)
from .core import (
    DEFAULT_PAGE_SIZE,
    DEPTH_LIMIT,
    MAX_PAGE_SIZE,
    NoEntry,
    UnsupportedMediaTypes,
    WrongTypeForRoute,
    apply_search,
    construct_data_response,
    construct_entries_response,
    construct_resource,
    construct_revisions_response,
    get_websocket_envelope_formatter,
    json_or_msgpack,
    resolve_media_type,
)
from .dependencies import (
    block,
    expected_shape,
    get_entry,
    get_root_tree,
    offset_param,
    shape_param,
)
from .file_response_with_range import FileResponseWithRange
from .links import links_for_node
from .settings import Settings, get_settings
from .utils import (
    filter_for_access,
    get_base_url,
    get_base_url_websocket,
    record_timing,
)

T = TypeVar("T")


def _patch_route_signature(
    query_registry: QueryRegistry,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """
    This is done dynamically at router startup.

    We check the registry of known search query types, which is user
    configurable, and use that to define the allowed HTTP query parameters for
    this route.

    Take a route that accept unspecified search queries as **filters.
    Return a wrapped version of the route that has the supported
    search queries explicitly spelled out in the function signature.

    This has no change in the actual behavior of the function,
    but it enables FastAPI to generate good OpenAPI documentation
    showing the supported search queries.

    """

    def inner(route: Callable[..., T]) -> Callable[..., T]:
        # Build a wrapper so that we can modify the signature
        # without mutating the wrapped original.

        async def route_with_sig(*args, **kwargs):
            return await route(*args, **kwargs)

        # Black magic here! FastAPI bases its validation and auto-generated swagger
        # documentation on the signature of the route function. We do not know what
        # that signature should be at compile-time. We only know it once we have a
        # chance to check the user-configurable registry of query types. Therefore,
        # we modify the signature here, at runtime, just before handing it to
        # FastAPI in the usual way.

        # When FastAPI calls the function with these added parameters, they will be
        # accepted via **filters.

        # Make a copy of the original parameters.
        signature = inspect.signature(route)
        parameters = list(signature.parameters.values())
        # Drop the **filters parameter from the signature.
        del parameters[-1]
        # Add a parameter for each field in each type of query.
        for name, query in query_registry.name_to_query_type.items():
            for field in dataclasses.fields(query):
                # The structured "alias" here is based on
                # https://mglaman.dev/blog/using-json-router-query-your-search-router-indexes
                if getattr(field.type, "__origin__", None) is list:
                    field_type = str
                else:
                    field_type = field.type
                injected_parameter = inspect.Parameter(
                    name=f"filter___{name}___{field.name}",
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=Query(
                        None, alias=f"filter[{name}][condition][{field.name}]"
                    ),
                    annotation=Optional[List[field_type]],
                )
                parameters.append(injected_parameter)
        route_with_sig.__signature__ = signature.replace(parameters=parameters)
        # End black magic

        return route_with_sig

    return inner


def get_router(
    query_registry: QueryRegistry,
    serialization_registry: SerializationRegistry,
    deserialization_registry: SerializationRegistry,
    validation_registry: ValidationRegistry,
    authenticators: dict[str, Union[ExternalAuthenticator, InternalAuthenticator]],
) -> APIRouter:
    router = APIRouter()

    @router.get("/", response_model=About)
    async def about(
        request: Request,
        settings: Settings = Depends(get_settings),
    ):
        # TODO The lazy import of entry modules and serializers means that the
        # lists of formats are not populated until they are first used. Not very
        # helpful for discovery! The registration can be made non-lazy, while the
        # imports of the underlying I/O libraries themselves (openpyxl, pillow,
        # etc.) can remain lazy.
        request.state.endpoint = "about"
        base_url = get_base_url(request)
        authentication = {
            "required": not settings.allow_anonymous_access,
        }
        provider_specs = []
        user_agent = request.headers.get("user-agent", "")
        # The name of the "internal" mode used to be "password".
        # This ensures back-compat with older Python clients.
        internal_mode_name = "internal"
        MINIMUM_INTERNAL_PYTHON_CLIENT_VERSION = packaging.version.parse("0.1.0b17")
        if user_agent.startswith("python-tiled/"):
            agent, _, raw_version = user_agent.partition("/")
            try:
                parsed_version = packaging.version.parse(raw_version)
            except Exception:
                pass
            else:
                if parsed_version < MINIMUM_INTERNAL_PYTHON_CLIENT_VERSION:
                    internal_mode_name = "password"
        for provider, authenticator in authenticators.items():
            if isinstance(authenticator, InternalAuthenticator):
                spec = {
                    "provider": provider,
                    "mode": internal_mode_name,
                    "links": {
                        "auth_endpoint": f"{base_url}/auth/provider/{provider}/token"
                    },
                    "confirmation_message": getattr(
                        authenticator, "confirmation_message", None
                    ),
                }
            elif isinstance(authenticator, ExternalAuthenticator):
                spec = {
                    "provider": provider,
                    "mode": "external",
                    "links": {
                        "auth_endpoint": f"{base_url}/auth/provider/{provider}/authorize"
                    },
                    "confirmation_message": getattr(
                        authenticator, "confirmation_message", None
                    ),
                }
            else:
                # It should be impossible to reach here.
                assert False
            provider_specs.append(spec)
        if provider_specs:
            # If there are *any* authenticaiton providers, these
            # endpoints will be added.
            authentication["links"] = {
                "whoami": f"{base_url}/auth/whoami",
                "apikey": f"{base_url}/auth/apikey",
                "refresh_session": f"{base_url}/auth/session/refresh",
                "revoke_session": f"{base_url}/auth/session/revoke/{{session_id}}",
                "logout": f"{base_url}/auth/logout",
            }
        authentication["providers"] = provider_specs

        return json_or_msgpack(
            request,
            About(
                library_version=__version__,
                api_version=0,
                formats={
                    structure_family: list(
                        serialization_registry.media_types(structure_family)
                    )
                    for structure_family in serialization_registry.structure_families
                },
                aliases={
                    structure_family: serialization_registry.aliases(structure_family)
                    for structure_family in serialization_registry.structure_families
                },
                queries=list(query_registry.name_to_query_type),
                authentication=authentication,
                links={
                    "self": base_url,
                    "documentation": f"{base_url}/docs",
                },
                meta={"root_path": request.scope.get("root_path") or "" + "/api"},
            ).model_dump(),
            expires=datetime.now(timezone.utc) + timedelta(seconds=600),
        )

    @router.get(
        "/search/{path:path}",
        response_model=schemas.Response[
            List[schemas.Resource[schemas.NodeAttributes, dict, dict]],
            schemas.PaginationLinks,
            dict,
        ],
    )
    @_patch_route_signature(query_registry)
    async def search(
        request: Request,
        path: str,
        fields: Optional[List[schemas.EntryFields]] = Query(list(schemas.EntryFields)),
        select_metadata: Optional[str] = Query(None),
        offset: Optional[int] = Query(0, alias="page[offset]", ge=0),
        limit: Optional[int] = Query(
            DEFAULT_PAGE_SIZE, alias="page[limit]", ge=0, le=MAX_PAGE_SIZE
        ),
        sort: Optional[str] = Query(None),
        max_depth: Optional[int] = Query(None, ge=0, le=DEPTH_LIMIT),
        omit_links: bool = Query(False),
        include_data_sources: bool = Query(False),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        settings: Settings = Depends(get_settings),
        _=Security(check_scopes, scopes=["read:metadata"]),
        **filters,
    ):
        entry = await get_entry(
            path,
            ["read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.container},
            getattr(request.app.state, "access_policy", None),
        )
        request.state.endpoint = "search"
        try:
            (
                resource,
                metadata_stale_at,
                must_revalidate,
            ) = await construct_entries_response(
                query_registry,
                entry,
                "/search",
                path,
                offset,
                limit,
                fields,
                select_metadata,
                omit_links,
                include_data_sources,
                filters,
                sort,
                get_base_url(request),
                resolve_media_type(request),
                max_depth=max_depth,
                exact_count_limit=settings.exact_count_limit,
            )
            # We only get one Expires header, so if different parts
            # of this response become stale at different times, we
            # cite the earliest one.
            entries_stale_at = getattr(entry, "entries_stale_at", None)
            headers = {}
            if (metadata_stale_at is None) or (entries_stale_at is None):
                expires = None
            else:
                expires = min(metadata_stale_at, entries_stale_at)
            if must_revalidate:
                headers["Cache-Control"] = "must-revalidate"
            return json_or_msgpack(
                request,
                resource.model_dump(),
                expires=expires,
                headers=headers,
            )
        except BrokenLink as err:
            raise HTTPException(status_code=HTTP_410_GONE, detail=err.args[0])
        except NoEntry:
            raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="No such entry.")
        except WrongTypeForRoute as err:
            raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail=err.args[0])
        except JMESPathError as err:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=f"Malformed 'select_metadata' parameter raised JMESPathError: {err}",
            )

    @router.get(
        "/distinct/{path:path}",
        response_model=schemas.GetDistinctResponse,
    )
    @_patch_route_signature(query_registry)
    async def distinct(
        path: str,
        request: Request,
        structure_families: bool = False,
        specs: bool = False,
        metadata: Optional[List[str]] = Query(default=[]),
        counts: bool = False,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:metadata"]),
        **filters,
    ):
        entry = await get_entry(
            path,
            ["read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if hasattr(entry, "get_distinct"):
            filtered = await apply_search(entry, filters, query_registry)
            distinct = await ensure_awaitable(
                filtered.get_distinct, metadata, structure_families, specs, counts
            )

            return json_or_msgpack(
                request,
                schemas.GetDistinctResponse.model_validate(distinct).model_dump(),
            )
        else:
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support distinct.",
            )

    @router.get(
        "/metadata/{path:path}",
        response_model=schemas.Response[
            schemas.Resource[schemas.NodeAttributes, dict, dict], dict, dict
        ],
    )
    async def metadata(
        request: Request,
        path: str,
        fields: Optional[List[schemas.EntryFields]] = Query(list(schemas.EntryFields)),
        select_metadata: Optional[str] = Query(None),
        max_depth: Optional[int] = Query(None, ge=0, le=DEPTH_LIMIT),
        omit_links: bool = Query(False),
        include_data_sources: bool = Query(False),
        root_path: bool = Query(False),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        settings: Settings = Depends(get_settings),
        _=Security(check_scopes, scopes=["read:metadata"]),
    ):
        """Fetch the metadata and structure information for one entry"""
        entry = await get_entry(
            path,
            ["read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        request.state.endpoint = "metadata"
        base_url = get_base_url(request)
        path_parts = [segment for segment in path.split("/") if segment]
        try:
            resource = await construct_resource(
                base_url,
                path_parts,
                entry,
                fields,
                select_metadata,
                omit_links,
                include_data_sources,
                resolve_media_type(request),
                max_depth=max_depth,
                exact_count_limit=settings.exact_count_limit,
            )
        except BrokenLink as err:
            raise HTTPException(status_code=HTTP_410_GONE, detail=err.args[0])
        except JMESPathError as err:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=f"Malformed 'select_metadata' parameter raised JMESPathError: {err}",
            )
        meta = {"root_path": request.scope.get("root_path") or "/"} if root_path else {}

        return json_or_msgpack(
            request,
            schemas.Response(data=resource, meta=meta).model_dump(),
            expires=getattr(entry, "metadata_stale_at", None),
        )

    @router.get(
        "/array/block/{path:path}", response_model=schemas.Response, name="array block"
    )
    async def array_block(
        request: Request,
        path: str,
        block=Depends(block),
        slice=Depends(NDSlice.from_query),
        expected_shape=Depends(expected_shape),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch a chunk of array-like data.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.array, StructureFamily.sparse},
            getattr(request.app.state, "access_policy", None),
        )
        shape = entry.structure().shape
        # Check that block dimensionality matches array dimensionality.
        ndim = len(shape)
        if len(block) != ndim:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Block parameter must have {ndim} comma-separated parameters, "
                    f"corresponding to the dimensions of this {ndim}-dimensional array."
                ),
            )
        if block == ():
            # Handle special case of numpy scalar.
            if shape != ():
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail=f"Requested scalar but shape is {entry.structure().shape}",
                )
            with record_timing(request.state.metrics, "read"):
                array = await ensure_awaitable(entry.read)
        else:
            try:
                with record_timing(request.state.metrics, "read"):
                    array = await ensure_awaitable(entry.read_block, block, slice)
            except IndexError:
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST, detail="Block index out of range"
                )
            if (expected_shape is not None) and (expected_shape != array.shape):
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail=f"The expected_shape {expected_shape} does not match the actual shape {array.shape}",
                )
        if array.nbytes > settings.response_bytesize_limit:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Response would exceed {settings.response_bytesize_limit}. "
                    "Use slicing ('?slice=...') to request smaller chunks."
                ),
            )
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    entry.structure_family,
                    serialization_registry,
                    array,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                )
        except UnsupportedMediaTypes as err:
            # raise HTTPException(status_code=406, detail=", ".join(err.supported))
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.get(
        "/array/full/{path:path}", response_model=schemas.Response, name="full array"
    )
    async def array_full(
        path: str,
        request: Request,
        slice=Depends(NDSlice.from_query),
        expected_shape=Depends(expected_shape),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch a slice of array-like data.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.array, StructureFamily.sparse},
            getattr(request.app.state, "access_policy", None),
        )
        structure_family = entry.structure_family
        # Deferred import because this is not a required dependency of the server
        # for some use cases.
        import numpy

        try:
            with record_timing(request.state.metrics, "read"):
                array = await ensure_awaitable(entry.read, slice)
            if structure_family == StructureFamily.array:
                array = numpy.asarray(array)  # Force dask or PIMS or ... to do I/O.
        except IndexError:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST, detail="Block index out of range"
            )
        if (expected_shape is not None) and (expected_shape != array.shape):
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=f"The expected_shape {expected_shape} does not match the actual shape {array.shape}",
            )
        if array.nbytes > settings.response_bytesize_limit:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Response would exceed {settings.response_bytesize_limit}. "
                    "Use slicing ('?slice=...') to request smaller chunks."
                ),
            )
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    structure_family,
                    serialization_registry,
                    array,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                )
        except UnsupportedMediaTypes as err:
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.delete("/stream/close/{path:path}")
    async def close_stream(
        request: Request,
        path: str,
        principal: Optional[schemas.Principal] = Depends(get_current_principal),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.array, StructureFamily.sparse},
            getattr(request.app.state, "access_policy", None),
        )
        await entry.close_stream()

    @router.websocket("/stream/single/{path:path}")
    async def websocket_endpoint(
        websocket: WebSocket,
        path: str,
        envelope_format: schemas.EnvelopeFormat = schemas.EnvelopeFormat.json,
        start: Optional[int] = None,
        principal: Optional[schemas.Principal] = Depends(
            get_current_principal_websocket
        ),
        authn_access_tags: Optional[Set[str]] = Depends(
            get_current_access_tags_websocket
        ),
        authn_scopes: Scopes = Depends(get_current_scopes_websocket),
    ):
        root_tree = websocket.app.state.root_tree
        websocket.state.metrics = collections.defaultdict(
            lambda: collections.defaultdict(lambda: 0)
        )
        entry = await get_entry(
            path,
            ["read:data", "read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            {},  # session_state,
            websocket.state.metrics,
            {
                StructureFamily.array,
                StructureFamily.container,
                StructureFamily.sparse,
            },
            getattr(websocket.app.state, "access_policy", None),
        )
        formatter = get_websocket_envelope_formatter(
            envelope_format, entry, deserialization_registry
        )
        base_websocket_url = URL(get_base_url_websocket(websocket))
        scheme = "https" if base_websocket_url.scheme == "wss" else "http"
        path_parts = [segment for segment in path.split("/") if segment]
        path_str = "/".join(path_parts)
        uri = f"{base_websocket_url.replace(scheme=scheme)}/array/full/{path_str}"
        handler = entry.make_ws_handler(websocket, formatter, uri)
        await handler(start)

    @router.get(
        "/table/partition/{path:path}",
        response_model=schemas.Response,
        name="table partition",
    )
    async def get_table_partition(
        path: str,
        request: Request,
        partition: int,
        column: Optional[List[str]] = Query(None, min_length=1),
        field: Optional[List[str]] = Query(None, min_length=1, deprecated=True),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch a partition (continuous block of rows) from a DataFrame [GET route].
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.table},
            getattr(request.app.state, "access_policy", None),
        )
        if (field is not None) and (column is not None):
            redundant_field_and_column = " ".join(
                (
                    "Cannot accept both 'column' and 'field' query parameters",
                    "in the same /table/partition request.",
                    "Include these query values using only the 'column' parameter.",
                )
            )
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST, detail=redundant_field_and_column
            )
        elif field is not None:
            field_is_deprecated = " ".join(
                (
                    "Query parameter 'field' is deprecated for the /table/partition route.",
                    "Instead use the query parameter 'column'.",
                )
            )
            warnings.warn(field_is_deprecated, DeprecationWarning)
        return await table_partition(
            request=request,
            partition=partition,
            entry=entry,
            column=(column or field),
            format=format,
            filename=filename,
            settings=settings,
        )

    @router.post(
        "/table/partition/{path:path}",
        response_model=schemas.Response,
        name="table partition",
    )
    async def post_table_partition(
        path: str,
        request: Request,
        partition: int,
        column: Optional[List[str]] = Body(None, min_length=1),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch a partition (continuous block of rows) from a DataFrame [POST route].
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.table},
            getattr(request.app.state, "access_policy", None),
        )
        return await table_partition(
            request=request,
            partition=partition,
            entry=entry,
            column=column,
            format=format,
            filename=filename,
            settings=settings,
        )

    async def table_partition(
        request: Request,
        partition: int,
        entry: AnyAdapter,
        column: Optional[List[str]],
        format: Optional[str],
        filename: Optional[str],
        settings: Settings,
    ):
        """
        Fetch a partition (continuous block of rows) from a DataFrame.
        """
        try:
            # The singular/plural mismatch here of "fields" and "field" is
            # due to the ?field=A&field=B&field=C... encodes in a URL.
            with record_timing(request.state.metrics, "read"):
                df = await ensure_awaitable(entry.read_partition, partition, column)
        except IndexError:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST, detail="Partition out of range"
            )
        except KeyError as err:
            (key,) = err.args
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST, detail=f"No such field {key}."
            )
        if df.memory_usage().sum() > settings.response_bytesize_limit:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Response would exceed {settings.response_bytesize_limit}. "
                    "Select a subset of the columns ('?field=...') to "
                    "request a smaller chunks."
                ),
            )
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    StructureFamily.table,
                    serialization_registry,
                    df,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                )
        except UnsupportedMediaTypes as err:
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.get(
        "/table/full/{path:path}",
        response_model=schemas.Response,
        name="full 'table' data",
    )
    async def get_table_full(
        request: Request,
        path: str,
        column: Optional[List[str]] = Query(None, min_length=1),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch the data for the given table [GET route].
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.table},
            getattr(request.app.state, "access_policy", None),
        )
        return await table_full(
            request=request,
            entry=entry,
            column=column,
            format=format,
            filename=filename,
            settings=settings,
        )

    @router.post(
        "/table/full/{path:path}",
        response_model=schemas.Response,
        name="full 'table' data",
    )
    async def post_table_full(
        request: Request,
        path: str,
        column: Optional[List[str]] = Body(None, min_length=1),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch the data for the given table [POST route].
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.table},
            getattr(request.app.state, "access_policy", None),
        )
        return await table_full(
            request=request,
            entry=entry,
            column=column,
            format=format,
            filename=filename,
            settings=settings,
        )

    async def table_full(
        request: Request,
        entry: AnyAdapter,
        column: Optional[List[str]],
        format: Optional[str],
        filename: Optional[str],
        settings: Settings,
    ):
        """
        Fetch the data for the given table.
        """
        try:
            with record_timing(request.state.metrics, "read"):
                data = await ensure_awaitable(entry.read, column)
        except KeyError as err:
            (key,) = err.args
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST, detail=f"No such field {key}."
            )
        if data.memory_usage().sum() > settings.response_bytesize_limit:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Response would exceed {settings.response_bytesize_limit}. "
                    "Select a subset of the columns to "
                    "request a smaller chunks."
                ),
            )
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    entry.structure_family,
                    serialization_registry,
                    data,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                    filter_for_access=None,
                )
        except UnsupportedMediaTypes as err:
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.get(
        "/container/full/{path:path}",
        response_model=schemas.Response,
        name="full 'container' metadata and data",
    )
    async def get_container_full(
        request: Request,
        path: str,
        principal: Optional[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        field: Optional[List[str]] = Query(None, min_length=1),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch the data for the given container via a GET request.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.container},
            getattr(request.app.state, "access_policy", None),
        )
        return await container_full(
            request=request,
            entry=entry,
            principal=principal,
            authn_access_tags=authn_access_tags,
            authn_scopes=authn_scopes,
            field=field,
            format=format,
            filename=filename,
        )

    @router.post(
        "/container/full/{path:path}",
        response_model=schemas.Response,
        name="full 'container' metadata and data",
    )
    async def post_container_full(
        request: Request,
        path: str,
        principal: Optional[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        field: Optional[List[str]] = Body(None, min_length=1),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch the data for the given container via a POST request.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.container},
            getattr(request.app.state, "access_policy", None),
        )
        return await container_full(
            request=request,
            entry=entry,
            principal=principal,
            authn_access_tags=authn_access_tags,
            authn_scopes=authn_scopes,
            field=field,
            format=format,
            filename=filename,
        )

    async def container_full(
        request: Request,
        entry,
        principal: Optional[Principal],
        authn_access_tags: Optional[Set[str]],
        authn_scopes: Scopes,
        field: Optional[List[str]],
        format: Optional[str],
        filename: Optional[str],
    ):
        """
        Fetch the data for the given container.
        """
        try:
            with record_timing(request.state.metrics, "read"):
                data = await ensure_awaitable(entry.read, fields=field)
        except KeyError as err:
            (key,) = err.args
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST, detail=f"No such field {key}."
            )
        curried_filter = partial(
            filter_for_access,
            access_policy=request.app.state.access_policy,
            principal=principal,
            authn_access_tags=authn_access_tags,
            authn_scopes=authn_scopes,
            scopes=["read:data"],
            metrics=request.state.metrics,
        )
        # TODO Walk node to determine size before handing off to serializer.
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    entry.structure_family,
                    serialization_registry,
                    data,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                    filter_for_access=curried_filter,
                )
        except UnsupportedMediaTypes as err:
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.get(
        "/node/full/{path:path}",
        response_model=schemas.Response,
        name="full 'container' or 'table'",
        deprecated=True,
    )
    async def node_full(
        request: Request,
        path: str,
        principal: Optional[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        field: Optional[List[str]] = Query(None, min_length=1),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch the data below the given node.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.table, StructureFamily.container},
            getattr(request.app.state, "access_policy", None),
        )
        try:
            with record_timing(request.state.metrics, "read"):
                data = await ensure_awaitable(entry.read, field)
        except KeyError as err:
            (key,) = err.args
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST, detail=f"No such field {key}."
            )
        if (entry.structure_family == StructureFamily.table) and (
            data.memory_usage().sum() > settings.response_bytesize_limit
        ):
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Response would exceed {settings.response_bytesize_limit}. "
                    "Select a subset of the columns ('?field=...') to "
                    "request a smaller chunks."
                ),
            )
        if entry.structure_family == StructureFamily.container:
            curried_filter = partial(
                filter_for_access,
                access_policy=request.app.state.access_policy,
                principal=principal,
                authn_access_tags=authn_access_tags,
                authn_scopes=authn_scopes,
                scopes=["read:data"],
                metrics=request.state.metrics,
            )
        else:
            curried_filter = None
            # TODO Walk node to determine size before handing off to serializer.
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    entry.structure_family,
                    serialization_registry,
                    data,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                    filter_for_access=curried_filter,
                )
        except UnsupportedMediaTypes as err:
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.get(
        "/awkward/buffers/{path:path}",
        response_model=schemas.Response,
        name="AwkwardArray buffers",
    )
    async def get_awkward_buffers(
        request: Request,
        path: str,
        form_key: Optional[List[str]] = Query(None, min_length=1),
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch a slice of AwkwardArray data.

        Note that there is a POST route on this same path with equivalent functionality.
        HTTP caches tends to engage with GET but not POST, so that GET route may be
        preferred for that reason. However, HTTP clients, servers, and proxies
        typically impose a length limit on URLs. (The HTTP spec does not specify
        one, but this is a pragmatic measure.) For requests with large numbers of
        form_key parameters, POST may be the only option.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.awkward},
            getattr(request.app.state, "access_policy", None),
        )
        return await _awkward_buffers(
            request=request,
            entry=entry,
            form_key=form_key,
            format=format,
            filename=filename,
            settings=settings,
        )

    @router.post(
        "/awkward/buffers/{path:path}",
        response_model=schemas.Response,
        name="AwkwardArray buffers",
    )
    async def post_awkward_buffers(
        request: Request,
        path: str,
        body: List[str],
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch a slice of AwkwardArray data.

        Note that there is a GET route on this same path with equivalent functionality.
        HTTP caches tends to engage with GET but not POST, so that GET route may be
        preferred for that reason. However, HTTP clients, servers, and proxies
        typically impose a length limit on URLs. (The HTTP spec does not specify
        one, but this is a pragmatic measure.) For requests with large numbers of
        form_key parameters, POST may be the only option.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.awkward},
            getattr(request.app.state, "access_policy", None),
        )
        return await _awkward_buffers(
            request=request,
            entry=entry,
            form_key=body,
            format=format,
            filename=filename,
            settings=settings,
        )

    async def _awkward_buffers(
        request: Request,
        entry,
        form_key: Optional[List[str]],
        format: Optional[str],
        filename: Optional[str],
        settings: Settings,
    ):
        structure_family = entry.structure_family
        structure = entry.structure()
        with record_timing(request.state.metrics, "read"):
            # The plural vs. singular mismatch is due to the way query parameters
            # are given as ?form_key=A&form_key=B&form_key=C.
            container = await ensure_awaitable(entry.read_buffers, form_key)
        if (
            sum(len(buffer) for buffer in container.values())
            > settings.response_bytesize_limit
        ):
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Response would exceed {settings.response_bytesize_limit}. "
                    "Use slicing ('?slice=...') to request smaller chunks."
                ),
            )
        components = (structure.form, structure.length, container)
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    structure_family,
                    serialization_registry,
                    components,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                )
        except UnsupportedMediaTypes as err:
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.get(
        "/awkward/full/{path:path}",
        response_model=schemas.Response,
        name="Full AwkwardArray",
    )
    async def awkward_full(
        request: Request,
        path: str,
        format: Optional[str] = None,
        filename: Optional[str] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        """
        Fetch a slice of AwkwardArray data.
        """
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.awkward},
            getattr(request.app.state, "access_policy", None),
        )
        structure_family = entry.structure_family
        # Deferred import because this is not a required dependency of the server
        # for some use cases.
        import awkward

        with record_timing(request.state.metrics, "read"):
            container = await ensure_awaitable(entry.read)
        structure = entry.structure()
        components = (structure.form, structure.length, container)
        array = awkward.from_buffers(*components)
        if array.nbytes > settings.response_bytesize_limit:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=(
                    f"Response would exceed {settings.response_bytesize_limit}. "
                    "Use slicing ('?slice=...') to request smaller chunks."
                ),
            )
        try:
            with record_timing(request.state.metrics, "pack"):
                return await construct_data_response(
                    structure_family,
                    serialization_registry,
                    components,
                    entry.metadata(),
                    request,
                    format,
                    specs=getattr(entry, "specs", []),
                    expires=getattr(entry, "content_stale_at", None),
                    filename=filename,
                )
        except UnsupportedMediaTypes as err:
            raise HTTPException(status_code=HTTP_406_NOT_ACCEPTABLE, detail=err.args[0])

    @router.post("/metadata/{path:path}", response_model=schemas.PostMetadataResponse)
    async def post_metadata(
        request: Request,
        path: str,
        body: schemas.PostMetadataRequest,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        _=Security(check_scopes, scopes=["write:metadata", "create"]),
    ):
        entry = await get_entry(
            path,
            ["write:metadata", "create"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        for data_source in body.data_sources:
            if data_source.assets:
                raise HTTPException(
                    "Externally-managed assets cannot be registered "
                    "using POST /metadata/{path}. Use POST /register/{path} instead."
                )
        if body.data_sources and not getattr(entry, "writable", False):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail=f"Data cannot be written at the path {path}",
            )
        return await _create_node(
            request=request,
            path=path,
            body=body,
            settings=settings,
            entry=entry,
            principal=principal,
            authn_access_tags=authn_access_tags,
            authn_scopes=authn_scopes,
        )

    @router.post("/register/{path:path}", response_model=schemas.PostMetadataResponse)
    async def post_register(
        request: Request,
        path: str,
        body: schemas.PostMetadataRequest,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        _=Security(check_scopes, scopes=["write:metadata", "create", "register"]),
    ):
        entry = await get_entry(
            path,
            ["write:metadata", "create", "register"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        return await _create_node(
            request=request,
            path=path,
            body=body,
            settings=settings,
            entry=entry,
            principal=principal,
            authn_access_tags=authn_access_tags,
            authn_scopes=authn_scopes,
        )

    async def _create_node(
        request: Request,
        path: str,
        body: schemas.PostMetadataRequest,
        settings: Settings,
        entry,
        principal: Optional[Principal],
        authn_access_tags: Optional[Set[str]],
        authn_scopes: Scopes,
    ):
        metadata, structure_family, specs, access_blob = (
            body.metadata,
            body.structure_family,
            body.specs,
            body.access_blob,
        )
        if structure_family == StructureFamily.container:
            structure = None
        else:
            if len(body.data_sources) != 1:
                raise NotImplementedError
            structure = body.data_sources[0].structure

        key = body.id or entry.context.key_maker()
        metadata_modified, metadata = await validate_specs(
            specs=specs,
            metadata=metadata,
            entry=None,  # the node doesn't exist yet
            structure_family=structure_family,
            structure=structure,
            settings=settings,
        )

        if request.app.state.access_policy is not None and hasattr(
            request.app.state.access_policy, "init_node"
        ):
            try:
                (
                    access_blob_modified,
                    access_blob,
                ) = await request.app.state.access_policy.init_node(
                    principal, authn_access_tags, authn_scopes, access_blob=access_blob
                )
            except ValueError as e:
                raise HTTPException(
                    status_code=HTTP_403_FORBIDDEN,
                    detail=f"Access policy rejects the provided access blob.\n{e}",
                )
        else:
            access_blob_modified = access_blob != {}
            access_blob = {}

        node = await entry.create_node(
            metadata=body.metadata,
            structure_family=body.structure_family,
            key=key,
            specs=body.specs,
            data_sources=body.data_sources,
            access_blob=access_blob,
        )
        links = links_for_node(
            structure_family, structure, get_base_url(request), path + f"/{node.key}"
        )
        response_data = {
            "id": node.key,
            "links": links,
            "data_sources": [ds.model_dump() for ds in node.data_sources],
        }
        if metadata_modified:
            response_data["metadata"] = metadata
        if access_blob_modified:
            response_data["access_blob"] = access_blob

        return json_or_msgpack(request, response_data)

    @router.put("/data_source/{path:path}")
    async def put_data_source(
        request: Request,
        path: str,
        body: schemas.PutDataSourceRequest,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:metadata", "register"]),
    ):
        entry = await get_entry(
            path,
            ["write:metadata", "register"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        await entry.put_data_source(data_source=body.data_source, patch=body.patch)

    @router.delete("/metadata/{path:path}")
    async def delete(
        request: Request,
        path: str,
        recursive: Optional[bool] = Query(
            False, description="Delete children recursively"
        ),
        external_only: Optional[bool] = Query(
            True,
            description=(
                "Delete the node, but only if this would not "
                "affect any internally-managed data sources"
            ),
        ),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data", "write:metadata"]),
    ):
        entry = await get_entry(
            path,
            ["write:data", "write:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if hasattr(entry, "delete"):
            await entry.delete(recursive=recursive, external_only=external_only)
        else:
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support deletion.",
            )
        return json_or_msgpack(request, None)

    @router.put("/array/full/{path:path}")
    async def put_array_full(
        request: Request,
        path: str,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.array, StructureFamily.sparse},
            getattr(request.app.state, "access_policy", None),
        )
        body = await request.body()
        if not hasattr(entry, "write"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node cannot accept array data.",
            )
        media_type = request.headers["content-type"]
        if entry.structure_family == "array":
            deserializer = deserialization_registry.dispatch("array", media_type)
        elif entry.structure_family == "sparse":
            deserializer = deserialization_registry.dispatch("sparse", media_type)
        else:
            raise NotImplementedError(entry.structure_family)
        await ensure_awaitable(entry.write, media_type, deserializer, entry, body)
        return json_or_msgpack(request, None)

    @router.put("/array/block/{path:path}")
    async def put_array_block(
        request: Request,
        path: str,
        block=Depends(block),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.array, StructureFamily.sparse},
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "write_block"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node cannot accept array data.",
            )

        body = await request.body()
        media_type = request.headers["content-type"]
        deserializer = deserialization_registry.dispatch(
            entry.structure_family, media_type
        )
        await ensure_awaitable(
            entry.write_block, block, media_type, deserializer, entry, body
        )
        return json_or_msgpack(request, None)

    @router.patch("/array/full/{path:path}")
    async def patch_array_full(
        request: Request,
        path: str,
        offset=Depends(offset_param),
        shape=Depends(shape_param),
        extend: bool = False,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.array},
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "patch"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node cannot accept array data.",
            )

        body = await request.body()
        media_type = request.headers["content-type"]
        deserializer = deserialization_registry.dispatch("array", media_type)
        structure = await ensure_awaitable(
            entry.patch, shape, offset, extend, media_type, deserializer, entry, body
        )
        return json_or_msgpack(request, structure)

    @router.put("/table/full/{path:path}")
    @router.put("/node/full/{path:path}", deprecated=True)
    async def put_node_full(
        request: Request,
        path: str,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.table},
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "write"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support writing.",
            )
        body = await request.body()
        media_type = request.headers["content-type"]
        deserializer = deserialization_registry.dispatch(
            StructureFamily.table, media_type
        )
        data = await ensure_awaitable(deserializer, body)
        await ensure_awaitable(entry.write, data)
        return json_or_msgpack(request, None)

    @router.put("/table/partition/{path:path}")
    async def put_table_partition(
        partition: int,
        path: str,
        request: Request,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "write_partition"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not supporting writing a partition.",
            )
        body = await request.body()
        media_type = request.headers["content-type"]
        deserializer = deserialization_registry.dispatch(
            StructureFamily.table, media_type
        )
        data = await ensure_awaitable(deserializer, body)
        await ensure_awaitable(entry.write_partition, data, partition)
        return json_or_msgpack(request, None)

    @router.patch("/table/partition/{path:path}")
    async def patch_table_partition(
        partition: int,
        path: str,
        request: Request,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "write_partition"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not supporting writing a partition.",
            )
        body = await request.body()
        media_type = request.headers["content-type"]
        deserializer = deserialization_registry.dispatch(
            StructureFamily.table, media_type
        )
        data = await ensure_awaitable(deserializer, body)
        await ensure_awaitable(entry.append_partition, data, partition)
        return json_or_msgpack(request, None)

    @router.put("/awkward/full/{path:path}")
    async def put_awkward_full(
        request: Request,
        path: str,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:data"]),
    ):
        entry = await get_entry(
            path,
            ["write:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            {StructureFamily.awkward},
            getattr(request.app.state, "access_policy", None),
        )
        body = await request.body()
        if not hasattr(entry, "write"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node cannot be written to.",
            )
        media_type = request.headers["content-type"]
        deserializer = deserialization_registry.dispatch(
            StructureFamily.awkward, media_type
        )
        structure = entry.structure()
        data = await ensure_awaitable(
            deserializer, body, structure.form, structure.length
        )
        await ensure_awaitable(entry.write, data)
        return json_or_msgpack(request, None)

    @router.patch("/metadata/{path:path}", response_model=schemas.PatchMetadataResponse)
    async def patch_metadata(
        request: Request,
        path: str,
        body: schemas.PatchMetadataRequest,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        drop_revision: bool = False,
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        _=Security(check_scopes, scopes=["write:metadata"]),
    ):
        entry = await get_entry(
            path,
            ["write:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "replace_metadata"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support update of metadata.",
            )

        if body.content_type == patch_mimetypes.JSON_PATCH:
            metadata = apply_json_patch(entry.metadata(), (body.metadata or []))
            specs = apply_json_patch((entry.specs or []), (body.specs or []))
            access_blob = apply_json_patch(entry.access_blob, (body.access_blob or []))
        elif body.content_type == patch_mimetypes.MERGE_PATCH:
            metadata = apply_merge_patch(entry.metadata(), (body.metadata or {}))
            # body.specs = [] clears specs, as per json merge patch specification
            # but we treat body.specs = None as "no modifications"
            current_specs = entry.specs or []
            target_specs = current_specs if body.specs is None else body.specs
            specs = apply_merge_patch(current_specs, target_specs)
            # json_merge_patch applies merge in-place, which would
            # otherwise modify the in-memory node and prevent the
            # access policy from sanity checking the access blob.
            # make a copy so we can compare the node against the
            # proposed new access blob.
            entry_access_blob_copy = deepcopy(entry.access_blob)
            access_blob = apply_merge_patch(
                entry_access_blob_copy, (body.access_blob or [])
            )
        else:
            raise HTTPException(
                status_code=HTTP_406_NOT_ACCEPTABLE,
                detail=f"valid content types: {', '.join(patch_mimetypes)}",
            )

        # Manually validate limits that bypass pydantic validation via patch
        if len(specs) > schemas.MAX_ALLOWED_SPECS:
            raise HTTPException(
                status_code=HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"Update cannot result in more than {schemas.MAX_ALLOWED_SPECS} specs",
            )
        if len(specs) != len(set(specs)):
            raise HTTPException(
                status_code=HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Update cannot result in non-unique specs",
            )

        metadata_modified, metadata = await validate_specs(
            specs=specs,
            metadata=metadata,
            entry=entry,
            settings=settings,
        )

        if request.app.state.access_policy is not None and hasattr(
            request.app.state.access_policy, "modify_node"
        ):
            try:
                (
                    access_blob_modified,
                    access_blob,
                ) = await request.app.state.access_policy.modify_node(
                    entry, principal, authn_access_tags, authn_scopes, access_blob
                )
            except ValueError as e:
                raise HTTPException(
                    status_code=HTTP_403_FORBIDDEN,
                    detail=f"Access policy rejects the provided access blob.\n{e}",
                )
        else:
            # Cannot modify the access blob if there is no access policy
            access_blob_modified = access_blob != entry.access_blob
            access_blob = entry.access_blob

        await entry.replace_metadata(
            metadata=metadata,
            specs=specs,
            access_blob=access_blob,
            drop_revision=drop_revision,
        )

        response_data = {"id": entry.node.key}
        if metadata_modified:
            response_data["metadata"] = metadata
        if access_blob_modified:
            response_data["access_blob"] = access_blob
        return json_or_msgpack(request, response_data)

    @router.put("/metadata/{path:path}", response_model=schemas.PutMetadataResponse)
    async def put_metadata(
        request: Request,
        path: str,
        body: schemas.PutMetadataRequest,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        drop_revision: bool = False,
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        _=Security(check_scopes, scopes=["write:metadata"]),
    ):
        entry = await get_entry(
            path,
            ["write:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "replace_metadata"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support update of metadata.",
            )

        metadata, specs, access_blob = (
            body.metadata if body.metadata is not None else entry.metadata(),
            body.specs if body.specs is not None else entry.specs,
            body.access_blob if body.access_blob is not None else entry.access_blob,
        )

        metadata_modified, metadata = await validate_specs(
            specs=specs,
            metadata=metadata,
            entry=entry,
            settings=settings,
        )

        if request.app.state.access_policy is not None and hasattr(
            request.app.state.access_policy, "modify_node"
        ):
            try:
                (
                    access_blob_modified,
                    access_blob,
                ) = await request.app.state.access_policy.modify_node(
                    entry, principal, authn_access_tags, authn_scopes, access_blob
                )
            except ValueError as e:
                raise HTTPException(
                    status_code=HTTP_403_FORBIDDEN,
                    detail=f"Access policy rejects the provided access blob.\n{e}",
                )
        else:
            # Cannot modify the access blob if there is no access policy
            access_blob_modified = access_blob != entry.access_blob
            access_blob = entry.access_blob

        await entry.replace_metadata(
            metadata=metadata,
            specs=specs,
            access_blob=access_blob,
            drop_revision=drop_revision,
        )

        response_data = {"id": entry.node.key}
        if metadata_modified:
            response_data["metadata"] = metadata
        if access_blob_modified:
            response_data["access_blob"] = access_blob
        return json_or_msgpack(request, response_data)

    @router.get("/revisions/{path:path}")
    async def get_revisions(
        request: Request,
        path: str,
        offset: Optional[int] = Query(0, alias="page[offset]", ge=0),
        limit: Optional[int] = Query(
            DEFAULT_PAGE_SIZE, alias="page[limit]", ge=0, le=MAX_PAGE_SIZE
        ),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:metadata"]),
    ):
        entry = await get_entry(
            path,
            ["read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "revisions"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support revisions.",
            )

        base_url = get_base_url(request)
        resource = await construct_revisions_response(
            entry,
            base_url,
            "/revisions",
            path,
            offset,
            limit,
            resolve_media_type(request),
        )
        return json_or_msgpack(request, resource.model_dump())

    @router.delete("/revisions/{path:path}")
    async def delete_revision(
        request: Request,
        path: str,
        number: int,
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["write:metadata"]),
    ):
        entry = await get_entry(
            path,
            ["write:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )
        if not hasattr(entry, "revisions"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support a del request for revisions.",
            )

        await entry.delete_revision(number)
        return json_or_msgpack(request, None)

    # For simplicity of implementation, we support a restricted subset of the full
    # Range spec. This could be extended if the need arises.
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Range
    RANGE_HEADER_PATTERN = re.compile(r"^bytes=(\d+)-(\d+)$")

    @router.get("/asset/bytes/{path:path}")
    async def get_asset(
        request: Request,
        path: str,
        id: int,
        relative_path: Optional[Path] = None,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )  # TODO: Separate scope for assets?
        if not settings.expose_raw_assets:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN,
                detail=(
                    "This Tiled server is configured not to allow "
                    "downloading raw assets."
                ),
            )
        if not hasattr(entry, "asset_by_id"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support downloading assets.",
            )

        asset = await entry.asset_by_id(id)
        if asset is None:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND,
                detail=f"This node exists but it does not have an Asset with id {id}",
            )
        if asset.is_directory:
            if relative_path is None:
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail=(
                        "This asset is a directory. Must specify relative path, "
                        f"from manifest provided by /asset/manifest/...?id={id}"
                    ),
                )
            if relative_path.is_absolute():
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail="relative_path query parameter must be a *relative* path",
                )
        else:
            if relative_path is not None:
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail="This asset is not a directory. The relative_path query parameter must not be set.",
                )
        if not asset.data_uri.startswith("file:"):
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail="Only download assets stored as file:// is currently supported.",
            )
        path = path_from_uri(asset.data_uri)
        if relative_path is not None:
            # Be doubly sure that this is under the Asset's data_uri directory
            # and not sneakily escaping it.
            if not os.path.commonpath([path, path / relative_path]) != path:
                # This should not be possible.
                raise RuntimeError(
                    f"Refusing to serve {path / relative_path} because it is outside "
                    "of the Asset's directory"
                )
            full_path = path / relative_path
        else:
            full_path = path
        stat_result = await anyio.to_thread.run_sync(os.stat, full_path)
        filename = full_path.name
        if "range" in request.headers:
            range_header = request.headers["range"]
            match = RANGE_HEADER_PATTERN.match(range_header)
            if match is None:
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail=(
                        "Only a Range headers of the form 'bytes=start-end' are supported. "
                        f"Could not parse Range header: {range_header}",
                    ),
                )
            range = start, _ = (int(match.group(1)), int(match.group(2)))
            if start > stat_result.st_size:
                raise HTTPException(
                    status_code=HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
                    headers={"content-range": f"bytes */{stat_result.st_size}"},
                )
            status_code = HTTP_206_PARTIAL_CONTENT
        else:
            range = None
            status_code = HTTP_200_OK
        return FileResponseWithRange(
            full_path,
            stat_result=stat_result,
            status_code=status_code,
            headers={"Content-Disposition": f'attachment; filename="{filename}"'},
            range=range,
        )

    @router.get("/asset/manifest/{path:path}")
    async def get_asset_manifest(
        request: Request,
        path: str,
        id: int,
        settings: Settings = Depends(get_settings),
        principal: Optional[Principal] = Depends(get_current_principal),
        root_tree=Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        _=Security(check_scopes, scopes=["read:data"]),
    ):
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            request.state.metrics,
            None,
            getattr(request.app.state, "access_policy", None),
        )  # TODO: Separate scope for assets?
        if not settings.expose_raw_assets:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN,
                detail=(
                    "This Tiled server is configured not to allow "
                    "downloading raw assets."
                ),
            )
        if not hasattr(entry, "asset_by_id"):
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail="This node does not support downloading assets.",
            )

        asset = await entry.asset_by_id(id)
        if asset is None:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND,
                detail=f"This node exists but it does not have an Asset with id {id}",
            )
        if not asset.is_directory:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail="This asset is not a directory. There is no manifest.",
            )
        if not asset.data_uri.startswith("file:"):
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail="Only download assets stored as file:// is currently supported.",
            )
        path = path_from_uri(asset.data_uri)
        manifest = []
        # Walk the directory and any subdirectories. Aggregate a list of all the
        # files, given as paths relative to the directory root.
        for root, _directories, files in os.walk(path):
            manifest.extend(Path(root, file) for file in files)
        return json_or_msgpack(request, {"manifest": manifest})

    async def validate_specs(
        specs: List[Spec],
        metadata: dict,
        entry: Optional[AnyAdapter] = None,
        structure_family: Optional[StructureFamily] = None,
        structure: Optional[dict] = None,
        settings: Settings = Depends(get_settings),
    ):
        metadata_modified = False

        # Specs should be ordered from most specific/constrained to least.
        # Validate them in reverse order, with the least constrained spec first,
        # because it may do normalization that helps pass the more constrained one.
        # Known Issue:
        # When there is more than one spec, it's possible for the validator for
        # Spec 2 to make a modification that breaks the validation for Spec 1.
        # For now we leave it to the server maintainer to ensure that validators
        # won't step on each other in this way, but this may need revisiting.
        for spec in reversed(specs):
            if spec not in validation_registry:
                if settings.reject_undeclared_specs:
                    raise HTTPException(
                        status_code=HTTP_400_BAD_REQUEST,
                        detail=f"Unrecognized spec: {spec.name}",
                    )
            else:
                validator = validation_registry(spec)
                try:
                    result = await ensure_awaitable(
                        validator, spec, metadata, entry, structure_family, structure
                    )
                except ValidationError as e:
                    raise HTTPException(
                        status_code=HTTP_400_BAD_REQUEST,
                        detail=f"failed validation for the {spec.name} spec:\n{e}",
                    )
                if result is not None:
                    metadata_modified = True
                    metadata = result

        return metadata_modified, metadata

    return router



import asyncio
import contextlib
import tempfile

import awkward
import h5py
import numpy
import pandas
import pytest
import sparse
import tifffile

from tiled.catalog import in_memory
from tiled.client import Context, from_context
from tiled.client.register import register
from tiled.client.smoke import read
from tiled.client.sync import copy
from tiled.client.utils import ClientError
from tiled.queries import Key
from tiled.server.app import build_app


@contextlib.contextmanager
def client_factory(readable_storage=None):
    with tempfile.TemporaryDirectory() as tempdir:
        catalog = in_memory(
            writable_storage=str(tempdir), readable_storage=readable_storage
        )
        app = build_app(catalog)
        with Context.from_app(app) as context:
            client = from_context(context)
            yield client


def populate_external(client, tmp_path):
    "Populate a client with registered externally-managed data."
    subdir = tmp_path / "subdir"
    subdir.mkdir()
    # array
    image = numpy.ones((3, 5))
    for filepath in [tmp_path / "image.tiff", subdir / "nested_image.tiff"]:
        tifffile.imwrite(filepath, image)
    # table
    for filepath in [tmp_path / "table.csv", subdir / "nested_table.csv"]:
        with open(filepath, "w") as file:
            file.write("x,y\n1,2\n3,4\n")
    # container
    for filepath in [tmp_path / "group.h5", subdir / "group.h5"]:
        with h5py.File(filepath, "w") as file:
            g = file.create_group("g")
            g["data"] = numpy.arange(3)
    # Note: Tiled does not currently happen to support any formats that it
    # identifies as 'awkward' or 'sparse'. Potentially it could, and this
    # test could be expanded to include those examples.
    asyncio.run(register(client, tmp_path))


def populate_internal(client):
    "Populate a client with uploaded internally-managed data."
    # array
    client.write_array([1, 2, 3], key="a", metadata={"color": "red"}, specs=["alpha"])
    # table
    df = pandas.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]})
    client.write_dataframe(df, key="b", metadata={"color": "green"}, specs=["beta"])
    # awkward
    client.write_awkward(
        awkward.Array([1, [2, 3]]), key="d", metadata={"color": "red"}, specs=["alpha"]
    )
    # sparse
    coo = sparse.COO(
        coords=numpy.array([[2, 5]]), data=numpy.array([1.3, 7.5]), shape=(10,)
    )
    client.write_sparse(key="e", coords=coo.coords, data=coo.data, shape=coo.shape)

    # nested
    container = client.create_container("c")
    container.write_array(
        [1, 2, 3], key="A", metadata={"color": "red"}, specs=["alpha"]
    )
    container.write_dataframe(df, key="B", metadata={"color": "green"}, specs=["beta"])
    container.write_awkward(
        awkward.Array([1, [2, 3]]), key="D", metadata={"color": "red"}, specs=["alpha"]
    )
    container.write_sparse(key="E", coords=coo.coords, data=coo.data, shape=coo.shape)


def test_copy_internal():
    with client_factory() as dest:
        with client_factory() as source:
            populate_internal(source)
            copy(source, dest)
            assert list(source) == list(dest)
            assert list(source["c"]) == list(dest["c"])
            read(dest, strict=True)


def test_copy_skip_conflict():
    with client_factory() as dest:
        with client_factory() as source:
            populate_internal(source)
            copy(source, dest)
            copy(source, dest, on_conflict="skip")
            assert list(source) == list(dest)
            assert list(source["c"]) == list(dest["c"])
            read(dest, strict=True)


def test_copy_warn_conflict():
    with client_factory() as dest:
        with client_factory() as source:
            populate_internal(source)
            copy(source, dest)
            with pytest.warns(UserWarning):
                copy(source, dest, on_conflict="warn")


def test_copy_error_conflict():
    with client_factory() as dest:
        with client_factory() as source:
            populate_internal(source)
            copy(source, dest)
            with pytest.raises(ClientError):
                copy(source, dest)


def test_copy_external(tmp_path):
    with client_factory(readable_storage=[tmp_path]) as dest:
        with client_factory() as source:
            populate_external(source, tmp_path)
            copy(source, dest)
            assert list(source) == list(dest)
            assert list(source["subdir"]) == list(dest["subdir"])
            read(dest, strict=True)


def test_copy_search_results():
    with client_factory() as dest:
        with client_factory() as source:
            populate_internal(source)
            results = source.search(Key("color") == "red")
            copy(results, dest)
            assert list(results) == list(dest)


def test_copy_items():
    with client_factory() as dest:
        with client_factory() as source:
            populate_internal(source)
            select_items = source.items()[:2]
            copy(select_items, dest)
            assert [key for key, _ in select_items] == list(dest)


def test_copy_dict():
    with client_factory() as dest:
        with client_factory() as source:
            populate_internal(source)
            select_dict = dict(source.items()[:2])
            copy(select_dict, dest)
            assert list(select_dict) == list(dest)



"""Add node_id index in data_sources table

Revision ID: 7809873ea2c7
Revises: 9331ed94d6ac
Create Date: 2025-08-01 11:26:54.658601

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "7809873ea2c7"
down_revision = "9331ed94d6ac"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(
        "idx_data_sources_node_id", table_name="data_sources", columns=["node_id"]
    )


def downgrade():
    op.drop_index("idx_data_sources_node_id", table_name="data_sources")



import asyncio
import concurrent.futures
import sqlite3
import threading
import time
from contextlib import closing

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context, record_history
from ..client.cache import Cache, CachedResponse, ThreadingMode, with_thread_lock
from ..server.app import build_app

tree = MapAdapter(
    {
        f"arr{i:03}": ArrayAdapter.from_array(i * numpy.arange(3), metadata={"i": i})
        for i in range(30)
    },
    metadata={"t": 1},
)


@pytest.fixture
def client():
    app = build_app(tree)
    with Context.from_app(app, cache=Cache()) as context:
        yield from_context(context)


def test_cache(client, tmpdir):
    cache = client.context.cache
    before_count = cache.count()
    before_size = cache.size()

    # First time: not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    after_count = cache.count()
    after_size = cache.size()
    assert after_count > before_count
    assert after_size > before_size

    # Second time: cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert isinstance(response, CachedResponse)


def test_no_cache(client):
    client.context.cache = None

    # First time: not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)


def test_lru_eviction(client):
    # First time: not cached
    client.context.cache.capacity = 5000
    num_items = len(client)
    for i in range(num_items):
        client.values()[i]

    # Not all the items fit.
    # If this fails, perhaps the bytesize of an item has drifted.
    # Tweak the capacity.
    assert client.context.cache.count() < num_items

    # Most recently accessed: still cached
    with record_history() as h:
        client.values()[i]
    for response in h.responses:
        assert isinstance(response, CachedResponse)

    # Least recently accessed: has been evicted
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: cached
    with record_history() as h:
        client.metadata
    for response in h.responses:
        assert not isinstance(response, CachedResponse)


def test_item_too_large_to_store(client):
    client.context.cache.max_item_size = 10

    # First time: not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: still not cached
    with record_history() as h:
        list(client.keys())
    for response in h.responses:
        assert not isinstance(response, CachedResponse)


def test_readonly_cache(client):
    # Start with a writable cache.

    # First time: not cached
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: cached
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert isinstance(response, CachedResponse)

    orig_size = client.context.cache.size()

    # Now use the same file as readonly cache.
    filepath = client.context.cache.filepath
    ro_cache = client.context.cache = Cache(filepath, readonly=True)

    # Still cached (from before)
    with record_history() as h:
        client.values()[0]
    for response in h.responses:
        assert isinstance(response, CachedResponse)

    # Now look at something new...
    # First time: not cached
    with record_history() as h:
        client.values()[1]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # Second time: still not cached
    with record_history() as h:
        client.values()[1]
    for response in h.responses:
        assert not isinstance(response, CachedResponse)

    # And cache size has not changed
    assert ro_cache.size() == orig_size

    # Implementation detail: database connection is read-only,
    # for defense in depth.
    with pytest.raises(sqlite3.OperationalError):
        with closing(ro_cache._conn.cursor()) as cur:
            cur.execute("DELETE FROM responses")


def test_clear_cache(client):
    cache = client.context.cache
    client.values()[0]
    assert cache.size() > 0
    assert cache.count() > 0
    cache.clear()
    assert cache.size() == cache.count() == 0


def test_not_thread_safe(client, monkeypatch):
    # Check that writes fail if thread safety is disabled
    monkeypatch.setattr(sqlite3, "threadsafety", ThreadingMode.SINGLE_THREAD)
    cache = client.context.cache
    # Clear the cache in another thread
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        future = executor.submit(cache.clear)
        with pytest.raises(RuntimeError):
            future.result(timeout=1)


@pytest.mark.skipif(
    sqlite3.threadsafety != ThreadingMode.SERIALIZED,
    reason="sqlite not built with thread safe support",
)
def test_thread_safety(client):
    cache = client.context.cache
    # Clear the cache in another thread
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        future = executor.submit(cache.clear)
        future.result(timeout=1)


@pytest.mark.asyncio
async def test_thread_lock():
    """Check that we can prevent concurrent thread writes."""

    class Timer:
        _lock = threading.Lock()
        sleep_time = 0.01

        @with_thread_lock
        def sleep(self):
            time.sleep(self.sleep_time)

    timer = Timer()
    # Run the timer twice concurrently
    loop = asyncio.get_running_loop()
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        coros = [
            loop.run_in_executor(executor, timer.sleep),
            loop.run_in_executor(executor, timer.sleep),
        ]
        t0 = time.perf_counter()
        await asyncio.gather(*coros)
        run_time = time.perf_counter() - t0
    # Check that the threads didn't run in parallel
    assert run_time >= (2.0 * timer.sleep_time), "Threads did not lock"



import sqlite3
import warnings
from contextlib import closing
from pathlib import Path
from sys import intern

import aiosqlite
import yaml

from ..utils import InterningLoader, ensure_specified_sql_driver


class AccessTagsParser:
    @classmethod
    def from_uri(cls, uri):
        if uri.startswith("file:"):
            uri = uri.split(":", 1)[1]
        uri = ensure_specified_sql_driver(uri)
        if not uri.startswith("sqlite+aiosqlite:"):
            raise ValueError(
                f"AccessTagsParser must be given a SQLite database URI, "
                f"i.e. 'sqlite:///...', 'sqlite+aiosqlite:///...'\n"
                f"Given URI results in: {uri=}"
            )
        uri_path = uri.split(":", 1)[1]
        if not uri_path.startswith("///"):
            raise ValueError(
                "Invalid URI provided, URI must contain 3 forward slashes, "
                "e.g. 'sqlite:///...'."
            )
        uri = f"file:{uri_path[3:]}"
        uri = uri if "?" in uri else f"{uri}?mode=ro"
        return cls(uri=uri)

    def __init__(self, db=None, uri=None):
        self._uri = uri
        self._db = db

    async def connect(self):
        if self._db is None:
            self._db = await aiosqlite.connect(
                self._uri, uri=True, check_same_thread=False
            )

    async def is_tag_defined(self, name):
        async with self._db.cursor() as cursor:
            await cursor.execute("SELECT 1 FROM tags WHERE name = ?;", (name,))
            row = await cursor.fetchone()
            found_tagname = bool(row)
        return found_tagname

    async def get_public_tags(self):
        async with self._db.cursor() as cursor:
            await cursor.execute("SELECT name FROM public_tags;")
            public_tags = {name for (name,) in await cursor.fetchall()}
        return public_tags

    async def get_scopes_from_tag(self, tagname, username):
        async with self._db.cursor() as cursor:
            await cursor.execute(
                "SELECT scope_name FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?;",
                (tagname, username),
            )
            user_tag_scopes = {scope for (scope,) in await cursor.fetchall()}
        return user_tag_scopes

    async def is_tag_owner(self, tagname, username):
        async with self._db.cursor() as cursor:
            await cursor.execute(
                "SELECT 1 FROM user_tag_owners WHERE tag_name = ? AND user_name = ?;",
                (tagname, username),
            )
            row = await cursor.fetchone()
            found_owner = bool(row)
        return found_owner

    async def is_tag_public(self, name):
        async with self._db.cursor() as cursor:
            await cursor.execute("SELECT 1 FROM public_tags WHERE name = ?;", (name,))
            row = await cursor.fetchone()
            found_public = bool(row)
        return found_public

    async def get_tags_from_scope(self, scope, username):
        async with self._db.cursor() as cursor:
            await cursor.execute(
                "SELECT tag_name FROM user_tag_scopes WHERE user_name = ? AND scope_name = ?;",
                (username, scope),
            )
            user_scope_tags = {tag for (tag,) in await cursor.fetchall()}
        return user_scope_tags


def create_access_tags_tables(db):
    with closing(db.cursor()) as cursor:
        tables_setup_sql = """
PRAGMA journal_mode = WAL;
PRAGMA synchronous = NORMAL;
PRAGMA temp_store = MEMORY;
PRAGMA foreign_keys = ON;
BEGIN TRANSACTION;
CREATE TABLE IF NOT EXISTS tags (
  id        INTEGER PRIMARY KEY,
  name      TEXT    UNIQUE NOT NULL,
  is_public INTEGER NOT NULL DEFAULT 0
    CHECK (is_public IN (0,1))
);
CREATE TABLE IF NOT EXISTS users (
  id   INTEGER PRIMARY KEY,
  name TEXT    UNIQUE NOT NULL
);
CREATE TABLE IF NOT EXISTS scopes (
  id   INTEGER PRIMARY KEY,
  name TEXT    UNIQUE NOT NULL
);
CREATE TABLE IF NOT EXISTS tags_users_scopes (
  tag_id    INTEGER NOT NULL
    REFERENCES tags(id)
    ON UPDATE CASCADE
    ON DELETE CASCADE,
  user_id   INTEGER NOT NULL
    REFERENCES users(id)
    ON UPDATE CASCADE
    ON DELETE CASCADE,
  scope_id  INTEGER NOT NULL
    REFERENCES scopes(id)
    ON UPDATE CASCADE
    ON DELETE CASCADE,
  PRIMARY KEY (tag_id, user_id, scope_id)
);
CREATE TABLE IF NOT EXISTS tag_owners (
  tag_id   INTEGER NOT NULL
    REFERENCES tags(id)
    ON UPDATE CASCADE
    ON DELETE CASCADE,
  user_id  INTEGER NOT NULL
    REFERENCES users(id)
    ON UPDATE CASCADE
    ON DELETE CASCADE,
  PRIMARY KEY (tag_id, user_id)
);
CREATE INDEX IF NOT EXISTS idx_tags_is_public ON tags (is_public);
CREATE INDEX IF NOT EXISTS idx_tus_users_scopes ON tags_users_scopes (user_id, scope_id);
CREATE INDEX IF NOT EXISTS idx_tus_users_scopes_scopeid ON tags_users_scopes (scope_id);
CREATE INDEX IF NOT EXISTS idx_tag_owners ON tag_owners (user_id);
CREATE VIEW IF NOT EXISTS public_tags AS
  SELECT name
  FROM tags
  WHERE is_public = 1;
CREATE VIEW IF NOT EXISTS user_tag_scopes AS
  SELECT
    u.name AS user_name,
    t.name AS tag_name,
    s.name AS scope_name
  FROM tags_users_scopes tus
    JOIN users  u ON u.id = tus.user_id
    JOIN tags   t ON t.id = tus.tag_id
    JOIN scopes s ON s.id = tus.scope_id;
CREATE VIEW IF NOT EXISTS user_tag_owners AS
  SELECT
    u.name AS user_name,
    t.name AS tag_name
  FROM tag_owners towner
    JOIN users u ON u.id = towner.user_id
    JOIN tags  t ON t.id = towner.tag_id;
PRAGMA optimize;
"""
        cursor.executescript(tables_setup_sql)
        db.commit()


def update_access_tags_tables(db, scopes, tags, owners, public_tags):
    with closing(db.cursor()) as cursor:
        tables_stage_sql = """
BEGIN TRANSACTION;
CREATE TEMP TABLE IF NOT EXISTS stage_tags (
  id        INTEGER,
  name      TEXT    NOT NULL,
  is_public INTEGER NOT NULL
    CHECK (is_public IN (0,1))
);
CREATE TEMP TABLE IF NOT EXISTS stage_users (
  id   INTEGER,
  name TEXT NOT NULL
);
CREATE TEMP TABLE IF NOT EXISTS stage_scopes (
  id   INTEGER,
  name TEXT NOT NULL
);
CREATE TEMP TABLE IF NOT EXISTS stage_tags_users_scopes (
  tag_id   INTEGER NOT NULL,
  user_id  INTEGER NOT NULL,
  scope_id INTEGER NOT NULL
);
CREATE TEMP TABLE IF NOT EXISTS stage_tag_owners (
  tag_id  INTEGER NOT NULL,
  user_id INTEGER NOT NULL
);
"""
        cursor.executescript(tables_stage_sql)

        # put all items into staging
        all_tags = [(tag, 0) for tag in tags] + [(tag, 0) for tag in owners]
        all_public = [(tag,) for tag in public_tags]
        all_users = {(user,) for users in tags.values() for user in users}
        all_users.update({(user,) for users in owners.values() for user in users})
        all_scopes = [(scope,) for scope in scopes]
        cursor.executemany(
            "INSERT INTO stage_tags(name, is_public) VALUES (?,?);",
            all_tags,
        )
        cursor.executemany(
            "UPDATE stage_tags SET is_public = 1 WHERE name = (?);", all_public
        )
        cursor.executemany(
            "INSERT INTO stage_users(name) VALUES (?);",
            all_users,
        )
        cursor.executemany(
            "INSERT INTO stage_scopes(name) VALUES (?);",
            all_scopes,
        )

        # push item names and metadata from staging to prod
        # then pull back ID values from prod to staging
        # note that UPSERT should always have a WHERE clause
        #   to avoid ambiguity. See the SQLite docs section 2.2
        #   https://www.sqlite.org/lang_upsert.html
        stage_push_sql = """
BEGIN TRANSACTION;
INSERT INTO tags (name, is_public)
  SELECT name, is_public
  FROM stage_tags
  WHERE true
  ON CONFLICT (name)
  DO UPDATE SET is_public = excluded.is_public;
INSERT INTO users(name) SELECT name FROM stage_users WHERE true ON CONFLICT(name) DO NOTHING;
INSERT INTO scopes(name) SELECT name FROM stage_scopes WHERE true ON CONFLICT(name) DO NOTHING;
UPDATE stage_tags SET id = (SELECT id FROM tags WHERE tags.name = stage_tags.name);
UPDATE stage_users SET id = (SELECT id FROM users WHERE users.name = stage_users.name);
UPDATE stage_scopes SET id = (SELECT id FROM scopes WHERE scopes.name = stage_scopes.name);
"""
        cursor.executescript(stage_push_sql)

        # load db IDs for items into memory
        cursor.execute("SELECT id, name FROM stage_tags")
        tags_to_id = {intern(name): tag_id for (tag_id, name) in cursor.fetchall()}
        cursor.execute("SELECT id, name FROM stage_users")
        users_to_id = {intern(name): user_id for (user_id, name) in cursor.fetchall()}
        cursor.execute("SELECT id, name FROM stage_scopes")
        scopes_to_id = {
            intern(name): scope_id for (scope_id, name) in cursor.fetchall()
        }

        # flatten relationships and push to staging
        tags_users_scopes = [
            (tags_to_id[tag], users_to_id[user], scopes_to_id[scope])
            for tag, users in tags.items()
            for user, scopes in users.items()
            for scope in scopes
        ]
        tag_owners = [
            (tags_to_id[tag], users_to_id[user])
            for tag, users in owners.items()
            for user in users
        ]
        cursor.executemany(
            "INSERT INTO stage_tags_users_scopes(tag_id, user_id, scope_id) VALUES (?,?,?);",
            tags_users_scopes,
        )
        cursor.executemany(
            "INSERT INTO stage_tag_owners(tag_id, user_id) VALUES (?,?);", tag_owners
        )

        # delete outdated tags from prod and add updated relationships to db
        # finally, drop the staging tables
        # to-do: consider refactoring this to indvidual execute statements
        #        to avoid implicit pre-mature commit by executescript()
        upsert_delete_sql = """
BEGIN TRANSACTION;
DELETE from tags WHERE id NOT in (SELECT id FROM stage_tags);
DELETE from users WHERE id NOT in (SELECT id FROM stage_users);
DELETE from scopes WHERE id NOT in (SELECT id FROM stage_scopes);
INSERT INTO tags_users_scopes (tag_id, user_id, scope_id)
  SELECT tag_id, user_id, scope_id FROM stage_tags_users_scopes
  WHERE true
  ON CONFLICT (tag_id, user_id, scope_id) DO NOTHING;
DELETE FROM tags_users_scopes
  WHERE (tag_id, user_id, scope_id)
  NOT IN (SELECT tag_id, user_id, scope_id FROM stage_tags_users_scopes);
INSERT INTO tag_owners (tag_id, user_id)
  SELECT tag_id, user_id FROM stage_tag_owners
  WHERE true
  ON CONFLICT (tag_id, user_id) DO NOTHING;
DELETE FROM tag_owners
  WHERE (tag_id, user_id) NOT IN (SELECT tag_id, user_id FROM stage_tag_owners);
DROP TABLE IF EXISTS stage_tags;
DROP TABLE IF EXISTS stage_users;
DROP TABLE IF EXISTS stage_scopes;
DROP TABLE IF EXISTS stage_tags_users_scopes;
DROP TABLE IF EXISTS stage_tag_owners;
PRAGMA optimize;
"""
        cursor.executescript(upsert_delete_sql)
        db.commit()


class AccessTagsCompiler:
    _MAX_TAG_NESTING = 5

    def __init__(
        self,
        scopes,
        tag_config,
        tags_db,
        group_parser,
    ):
        self.scopes = scopes or {}
        self.tag_config = tag_config
        self.connection = sqlite3.connect(
            tags_db["uri"], uri=True, check_same_thread=False
        )
        self.group_parser = group_parser

        self.max_tag_nesting = max(self._MAX_TAG_NESTING, 0)
        self.public_tag = intern("public".casefold())
        self.invalid_tag_names = [name.casefold() for name in []]

        self.roles = {}
        self.tags = {}
        self.tag_owners = {}
        self.compiled_tags = {self.public_tag: {}}
        self.compiled_public = set({self.public_tag})
        self.compiled_tag_owners = {}

        create_access_tags_tables(self.connection)

    def load_tag_config(self):
        if isinstance(self.tag_config, str) or isinstance(self.tag_config, Path):
            try:
                with open(Path(self.tag_config)) as tag_config_file:
                    tag_definitions = yaml.load(tag_config_file, Loader=InterningLoader)
                    self.roles.update(tag_definitions.get("roles", {}))
                    self.tags.update(tag_definitions["tags"])
                    self.tag_owners.update(tag_definitions.get("tag_owners", {}))
            except FileNotFoundError as e:
                raise ValueError(
                    f"The tag config file {self.tag_config!s} doesn't exist."
                ) from e
        elif isinstance(self.tag_config, dict):
            tag_definitions = self.tag_config
            self.roles.update(tag_definitions.get("roles", {}))
            self.tags.update(tag_definitions["tags"])
            self.tag_owners.update(tag_definitions.get("tag_owners", {}))

    def _dfs(self, current_tag, tags, seen_tags, nested_level=0):
        if current_tag in self.compiled_tags:
            return self.compiled_tags[current_tag], current_tag in self.compiled_public
        if current_tag in seen_tags:
            return {}, False
        if nested_level > self.max_tag_nesting:
            raise RecursionError(
                f"Exceeded maximum tag nesting of {self.max_tag_nesting} levels"
            )

        public_auto_tag = False
        seen_tags.add(current_tag)
        users = {}
        for tag in tags[current_tag]:
            if tag.casefold() == self.public_tag:
                public_auto_tag = True
                continue
            try:
                child_users, child_public = self._dfs(
                    tag, tags, seen_tags, nested_level + 1
                )
                public_auto_tag = public_auto_tag or child_public
                users.update(child_users)
            except (RecursionError, ValueError) as e:
                raise RuntimeError(
                    f"Tag compilation failed at tag: {current_tag}"
                ) from e

        if public_auto_tag:
            self.compiled_public.add(current_tag)

        if "users" in self.tags[current_tag]:
            for user in self.tags[current_tag]["users"]:
                username = user["name"]
                if all(k in user for k in ("scopes", "role")):
                    raise ValueError(
                        f"Cannot define both 'scopes' and 'role' for a user. {username=}"
                    )
                elif not any(k in user for k in ("scopes", "role")):
                    raise ValueError(
                        f"Must define either 'scopes' or 'role' for a user. {username=}"
                    )

                user_scopes = set(
                    self.roles[user["role"]]["scopes"]
                    if ("role" in user) and (user["role"] in self.roles)
                    else user.get("scopes", [])
                )
                if not user_scopes:
                    raise ValueError(f"Scopes must not be empty. {username=}")
                if not user_scopes.issubset(self.scopes):
                    raise ValueError(
                        f"Scopes for {username=} are not in the valid set of scopes. The invalid scopes are:"
                        f"{user_scopes.difference(self.scopes)}"
                    )
                users.setdefault(username, set())
                users[username].update(user_scopes)

        if "groups" in self.tags[current_tag]:
            for group in self.tags[current_tag]["groups"]:
                groupname = group["name"]
                if all(k in group for k in ("scopes", "role")):
                    raise ValueError(
                        f"Cannot define both 'scopes' and 'role' for a group. {groupname=}"
                    )
                elif not any(k in group for k in ("scopes", "role")):
                    raise ValueError(
                        f"Must define either 'scopes' or 'role' for a group. {groupname=}"
                    )

                group_scopes = set(
                    self.roles[group["role"]]["scopes"]
                    if ("role" in group) and (group["role"] in self.roles)
                    else group.get("scopes", [])
                )
                if not group_scopes:
                    raise ValueError(f"Scopes must not be empty. {groupname=}")
                if not group_scopes.issubset(self.scopes):
                    raise ValueError(
                        f"Scopes for {groupname=} are not in the valid set of scopes. The invalid scopes are:"
                        f"{group_scopes.difference(self.scopes)}"
                    )

                try:
                    usernames = self.group_parser(groupname)
                except KeyError:
                    warnings.warn(
                        f"Group with {groupname=} does not exist - skipping",
                        UserWarning,
                    )
                    continue
                else:
                    for username in usernames:
                        username = intern(username)
                        users.setdefault(username, set())
                        users[username].update(group_scopes)

        self.compiled_tags[current_tag] = users
        return users, public_auto_tag

    def compile(self):
        for role in self.roles.values():
            if "scopes" not in role:
                raise ValueError(f"Scopes must be defined for a role. {role=}")
            if not role["scopes"]:
                raise ValueError(f"Scopes must not be empty. {role=}")
            if not set(role["scopes"]).issubset(self.scopes):
                raise ValueError(
                    f"Scopes for {role=} are not in the valid set of scopes. The invalid scopes are:"
                    f'{set(role["scopes"]).difference(self.scopes)}'
                )

        adjacent_tags = {}
        for tag, members in self.tags.items():
            if tag.casefold() == self.public_tag:
                raise ValueError(
                    f"'Public' tag '{self.public_tag}' cannot be redefined."
                )
            if tag.casefold() in self.invalid_tag_names:
                raise ValueError(
                    f"Tag 'tag' is an invalid tag name.\n"
                    f"The invalid tag names are: {self.invalid_tag_names}"
                )
            adjacent_tags[tag] = set()
            if "auto_tags" in members:
                for auto_tag in members["auto_tags"]:
                    if (
                        auto_tag["name"] not in self.tags
                        and auto_tag["name"].casefold() != self.public_tag
                    ):
                        raise KeyError(
                            f"Tag '{tag}' has nested tag '{auto_tag}' which does not have a definition."
                        )
                    adjacent_tags[tag].add(auto_tag["name"])

        for tag in adjacent_tags:
            try:
                self._dfs(tag, adjacent_tags, set())
            except (RecursionError, ValueError) as e:
                raise RuntimeError(f"Tag compilation failed at tag: {tag}") from e

        for tag in self.tag_owners:
            self.compiled_tag_owners.setdefault(tag, set())
            if "users" in self.tag_owners[tag]:
                for user in self.tag_owners[tag]["users"]:
                    username = user["name"]
                    self.compiled_tag_owners[tag].add(username)
            if "groups" in self.tag_owners[tag]:
                for group in self.tag_owners[tag]["groups"]:
                    groupname = group["name"]
                    try:
                        usernames = self.group_parser(groupname)
                    except KeyError:
                        warnings.warn(
                            f"Group with {groupname=} does not exist - skipping",
                            UserWarning,
                        )
                        continue
                    else:
                        for username in usernames:
                            username = intern(username)
                            self.compiled_tag_owners[tag].add(username)

        update_access_tags_tables(
            self.connection,
            self.scopes,
            self.compiled_tags,
            self.compiled_tag_owners,
            self.compiled_public,
        )

    def clear_raw_tags(self):
        self.roles = {}
        self.tags = {}
        self.tag_owners = {}

    def recompile(self):
        self.compiled_tags = {self.public_tag: {}}
        self.compiled_public = set({self.public_tag})
        self.compiled_tag_owners = {}
        self.compile()



from pathlib import Path

import h5py
import numpy as np
import pandas
import pytest
import zarr

from ..adapters.zarr import ZARR_LIB_V2
from ..catalog import in_memory
from ..client import Context, from_context, tree
from ..client.register import register
from ..server.app import build_app


@pytest.mark.asyncio
async def test_excel(tmpdir):
    df = pandas.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    df.to_excel(tmpdir / "spreadsheet.xlsx")
    catalog = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, tmpdir)
        tree(client)
        client["spreadsheet"]["Sheet1"].read()


@pytest.mark.asyncio
async def test_zarr_array(tmpdir):
    z = zarr.open(
        str(tmpdir / "za.zarr"), mode="w", shape=(3,), chunks=(3,), dtype="i4"
    )
    z[:] = [1, 2, 3]
    catalog = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, tmpdir)
        tree(client)
        client["za"].read()
    z.store.close()


@pytest.mark.asyncio
async def test_zarr_group(tmp_path: Path):
    root = zarr.open(tmp_path / "zg.zarr", mode="w")
    x_array = np.array([1, 2, 3])
    y_array = np.array([4, 5, 6])

    if ZARR_LIB_V2:
        root.create_dataset("x", data=x_array)
        root.create_dataset("y", data=y_array)
    else:
        root.create_array(name="x", data=x_array)
        root.create_array(name="y", data=y_array)

    catalog = in_memory(readable_storage=[tmp_path])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, tmp_path)
        tree(client)
        # Normally, zarr would have 'attributes' stored as internal dictionary in the
        # metadata, but HDF5 does not support nested dictionaries.
        client["zg"].replace_metadata(
            {"attributes": "", "zarr_format": 2 if ZARR_LIB_V2 else 3}
        )
        client["zg"].export(tmp_path / "stuff.h5")
        assert client["zg"]["x"].read().data == x_array
        assert client["zg"]["y"].read().data == y_array


@pytest.mark.asyncio
async def test_hdf5(tmpdir, buffer):
    with h5py.File(str(tmpdir / "h.h5"), "w") as file:
        file["x"] = [1, 2, 3]
        group = file.create_group("g")
        group["y"] = [4, 5, 6]
    catalog = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, tmpdir)
        tree(client)
        client["h"]["x"].read()
        client["h"]["g"]["y"].read()

        client.export(buffer, format="application/json")



import copy
import itertools
from typing import Any, List, Optional, Tuple, Union
from urllib.parse import quote_plus

import dask.base
import dask.dataframe
import numpy
import pandas
import sparse
from numpy._typing import NDArray

from ..adapters.array import slice_and_shape_from_block_and_chunks
from ..catalog.orm import Node
from ..ndslice import NDSlice
from ..storage import FileStorage, Storage
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource
from ..structures.sparse import COOStructure, SparseStructure
from ..type_aliases import JSON
from ..utils import path_from_uri
from .utils import init_adapter_from_catalog


def load_block(uri: str) -> Tuple[List[int], Tuple[NDArray[Any], Any]]:
    """

    Parameters
    ----------
    uri :

    Returns
    -------

    """
    # TODO This can be done without pandas.
    # Better to use a plain I/O library.
    df = pandas.read_parquet(path_from_uri(uri))
    coords = df[df.columns[:-1]].values.T
    data = df["data"].values
    return coords, data


class SparseBlocksParquetAdapter:
    """ """

    structure_family = StructureFamily.sparse
    supported_storage = {FileStorage}

    def __init__(
        self,
        data_uris: List[str],
        structure: COOStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        data_uris :
        structure :
        metadata :
        specs :
        """
        num_blocks = (range(len(n)) for n in structure.chunks)
        self.blocks = {}
        for block, uri in zip(itertools.product(*num_blocks), data_uris):
            self.blocks[block] = uri
        self._structure = structure
        self._metadata = metadata or {}
        self.specs = list(specs or [])

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[SparseStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "SparseBlocksParquetAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    @classmethod
    def init_storage(
        cls,
        storage: Storage,
        data_source: DataSource[COOStructure],
        path_parts: List[str],
    ) -> DataSource[COOStructure]:
        """

        Parameters
        ----------
        data_uri :
        structure :

        Returns
        -------

        """
        data_source = copy.deepcopy(data_source)  # Do not mutate caller input.
        data_uri = storage.uri + "".join(
            f"/{quote_plus(segment)}" for segment in path_parts
        )
        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)

        num_blocks = (range(len(n)) for n in data_source.structure.chunks)
        assets = [
            Asset(
                data_uri=f"{data_uri}/block-{'.'.join(map(str, block))}.parquet",
                is_directory=False,
                parameter="data_uris",
                num=i,
            )
            for i, block in enumerate(itertools.product(*num_blocks))
        ]
        data_source.assets.extend(assets)
        return data_source

    def metadata(self) -> JSON:
        """

        Returns
        -------

        """
        return self._metadata

    def write_block(
        self,
        data: Union[dask.dataframe.DataFrame, pandas.DataFrame],
        block: Tuple[int, ...],
        slice: NDSlice = NDSlice(...),
    ) -> None:
        if slice:
            raise NotImplementedError(
                "Writing into a slice of a sparse block is not yet supported."
            )
        "Write into a block of the array."
        uri = self.blocks[block]
        data.to_parquet(path_from_uri(uri))

    def write(self, data: Union[dask.dataframe.DataFrame, pandas.DataFrame]) -> None:
        """

        Parameters
        ----------
        data :

        Returns
        -------

        """
        if len(self.blocks) > 1:
            raise NotImplementedError
        uri = self.blocks[(0,) * len(self._structure.shape)]
        data.to_parquet(path_from_uri(uri))

    def read(self, slice: NDSlice = NDSlice(...)) -> sparse.COO:
        """

        Parameters
        ----------
        slice :

        Returns
        -------

        """
        all_coords = []
        all_data = []
        for block, uri in self.blocks.items():
            coords, data = load_block(uri)
            offsets = []
            for b, c in zip(block, self.structure().chunks):
                offset = sum(c[:b])
                offsets.append(offset)
            global_coords = coords + [[i] for i in offsets]
            all_coords.append(global_coords)
            all_data.append(data)
        arr = sparse.COO(
            data=numpy.concatenate(all_data),
            coords=numpy.concatenate(all_coords, axis=-1),
            shape=self._structure.shape,
        )
        return arr[slice] if slice else arr

    def read_block(
        self, block: Tuple[int, ...], slice: NDSlice = NDSlice(...)
    ) -> sparse.COO:
        """

        Parameters
        ----------
        block :
        slice :

        Returns
        -------

        """
        coords, data = load_block(self.blocks[block])
        _, shape = slice_and_shape_from_block_and_chunks(block, self._structure.chunks)
        arr = sparse.COO(data=data[:], coords=coords[:], shape=shape)
        return arr[slice] if slice else arr

    def structure(self) -> COOStructure:
        """

        Returns
        -------

        """
        return self._structure



import pytest
from httpx import ASGITransport, AsyncClient
from starlette.status import HTTP_200_OK

from ..server.app import build_app


@pytest.mark.parametrize("path", ["/", "/docs", "/healthz"])
@pytest.mark.asyncio
async def test_meta_routes(path):
    transport = ASGITransport(app=build_app({}))
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        response = await client.get(path)
    assert response.status_code == HTTP_200_OK
